from functools import reduce
from importlib.resources import files
import json
import sys

# from aldepyde.biomolecule_old import dna, rna, amino_acid, Residue

class _DataSingleton(type):
    _instance = {}
    def __call__(cls, *args, singleton=True, **kwargs):
        if singleton:
            return super(_DataSingleton, cls).__call__(*args, **kwargs)
        if cls not in cls._instance:
            cls._instance[cls] = super(_DataSingleton, cls).__call__(*args, **kwargs)
        return cls._instance[cls]


class Data(metaclass=_DataSingleton):
    # Map paths
    _map = ("map", ("map",))
    _dna_map = ("dna_map", ("map", "dna"))
    _rna_map = ("rna_map", ("map", "rna"))
    _amino_map = ("amino_map", ("map", "amino_acid"))

    def __init__(self, json_location = None):
        if json_location is None:
            ###### If something breaks, you can use this as a backup way to access the remode_data ######
            # base = os.path.dirname(os.path.abspath(__file__))
            # json_location = os.path.join(base, "data", "chemistry.data")
            json_location = files("aldepyde.data").joinpath("chemistry.data")
        self.json_location = json_location
        self._loaded = {}


    # Technically, this is the only function we need.
    # You get the rest because I care <3
    def load_values(self, *args, store_as: str = None):
        with open(self.json_location) as js:
            if args in self._loaded:
                return self.__dict__[self._loaded[args]]
            j_data = reduce(lambda d, key: d[key], args, json.load(js))
            if store_as is not None and args not in self._loaded:
                self._loaded[args] = store_as
                setattr(self, store_as, j_data)
                self.__dict__[store_as]['_key'] = args
            return j_data

    def unload(self, attr_name: str) -> bool:
        if attr_name not in self.__dict__.keys():
            return False
        try:
            item = self.__dict__.pop(attr_name)['_key']
            self._loaded.pop(item)
        except KeyError:
            # This really shouldn't occur unless you're trying
            raise KeyError(f'An error occured while attempting to remove {attr_name} from the data object.'
                     f' Are you sure you are attempting to unload a loaded value?')
        return True

    # TODO check if something is already loaded
    def GrabParent(self, *args):
        pass

    # Cute lil' recursive method that shows the structure of a loaded data. Maybe not so practical
    # at runtime, but helpful for debugging and planning your loads
    def reveal(self, *args, indent="  ") -> str:
        j_data = self.load_values(*args, store_as=None)
        return self._reveal_helper(j_data, indent, indent)

    def _reveal_helper(self, js: dict, indent, adder, ret_str="") -> str:
        for key in js:
            if not isinstance(js[key], dict):
                continue
            ret_str += indent+key + "\n"
            ret_str = self._reveal_helper(js[key], indent+adder, adder, ret_str)
        return ret_str

    def Map(self, residue: str|None, *args, store_as: str|None =_map[0], residue_type: str ='amino_acid') -> None|str:
        if args == ():
            args = self._map[1]
            if store_as is None:
                store_as = self._map[0]
        residue_type = residue_type.lower()
        if residue_type.lower() not in ["dna", "rna", "amino_acid", "element"]:
            print("Allowed residue_type mappins are 'dna', 'rna', 'amino_acid', and 'element'", file=sys.stderr)
        map = self.load_values(*args, store_as=store_as)
        if residue is None: # Just initialize self.map
            return None
        return map[residue_type][residue.lower()]


    # # These three could probably be condensed
    # def CheckDNA(self, value: str, *args) -> bool:
    #     if args == ():
    #         args = self._dna_map[1]
    #     map = self.load_values(*args, store_as=None)
    #     if value in map['dna'].keys():
    #         return True
    #     return False
    #
    #
    # def CheckRNA(self, value: str, *args) -> bool:
    #     if args == ():
    #         args = self._map[1]
    #     map = self.load_values(*args, store_as=None)
    #     if value in map['dna'].keys():
    #         return True
    #     return False
    #
    # def CheckAA(self, value: str, *args) -> bool:
    #     if args == ():
    #         args = self._map[1]
    #     map = self.load_values(*args, store_as=None)
    #     if value in map['dna'].keys():
    #         return True
    #     return False
    #
    # def CheckResidue(self, value: str, *args) -> bool:
    #     if args == ():
    #         args = self._map[1]
    #     if self.CheckAA(value, *args):
    #         return True
    #     if self.CheckDNA(value, *args):
    #         return True
    #     if self.CheckRNA(value, *args):
    #         return True
    #     return False
    #
    # # This method determines if something is DNA, RNA, or an amino acid.
    # # Don't be cheeky with this. If you aren't following the IUPAC naming schemes,
    # # you're gonna have a bad time.
    # #
    # # RNA has exclusively 1-letter codes: A, C, T, G, etc.
    # # DNA has exclusively 2-letter codes: DA, DC, DT, DG, etc.
    # # Amino acids have exclusively 3-letter codes
    # # def ExtrapolateResidueType(self, value: str) -> object:
    # #     if self.CheckRNA(value):
    # #         return rna
    # #     if self.CheckDNA(value):
    # #         return dna
    # #     if self.CheckAA(value):
    # #         return amino_acid
    # #     return Residue

data = Data()






from abc import ABC, abstractmethod
from typing import ClassVar, Set
import sys

# Define monomer classes and alphabets
class PolymerClassifier(ABC):
    def __init__(self, alphabet:str|list|set):
        self.alphabet: ClassVar[set[str]] = alphabet if isinstance(alphabet, set) else set(alphabet)
        # self.classes = classes
        # self.flags = flags

    def is_valid(self, residue:str) -> bool:
        return residue in self.alphabet

    def in_class(self, residue:str, class_name:str) -> bool:
        return residue in self.classes[class_name]

    def in_classes(self, residue:str) -> list:
        inside = list()
        for key in self.classes.keys():
            if residue in self.classes[key]:
                inside.append(key)
        return inside

    def get_classes(self) -> list:
        return list(self.classes.keys())

    def get_alphabet(self) -> set:
        return self.alphabet

    def load_alphabet(self, alphabet:str|list):
        self.alphabet = alphabet if isinstance(alphabet, list) else list(alphabet)

    def load_classes(self, classes:dict):
        self.classes = classes

    def get_class_members(self, class_name):
        return self.classes[class_name]

    def append_class(self, class_name:str, members:set|str):
        if isinstance(members, str):
            self.alphabet |= set(members)
            self.classes[class_name] |= set(members)
        else:
            self.alphabet |= members
            self.classes[class_name] |= members


    def add_class(self, class_name:str, members=None):
        if members is None:
            members = set("")
        if isinstance(members, str):
            members = set(members)
        if class_name not in self.classes.keys():
            self.classes[class_name] = set()
        self.append_class(class_name, members)

    def remove_class(self, class_name:str, clean_alphabet=False):
        members = self.classes[class_name]
        if self.classes.pop(class_name, None) is None:
            sys.stderr.write(f"Tried to remove class '{class_name},' which does not exist\n")
        else:
            if clean_alphabet:
                self.clean_alphabet(members)

    def clean_alphabet(self, members:set):
        orphans = set()
        for member in members:
            for cls in self.classes.keys():
                if member in self.classes[cls]:
                    break
            else:
                orphans.add(member)
        self.remove_from_alphabet(orphans)

    def remove_from_alphabet(self, members:set):
        self.alphabet -= members
        for key in self.classes.keys():
            self.classes[key] -= members

    @abstractmethod
    def set_default_alphabet(self):
        pass

    @abstractmethod
    def set_default_classes(self):
        pass

    # @abstractmethod
    # def default_flags(self):
    #     pass

    def reset_defaults(self):
        self.set_default_classes()
        self.set_default_alphabet()



# Submodule for reading, writing, and fetching PDB/mmcif files
from aldepyde.configurable import Configurable
from aldepyde.remode_data import RemoteFileHandler
from ._AtomFactory import *
from ._Atom import Atom
from aldepyde import cache
from enum import Enum
import urllib.request
import requests
import gzip
from io import BytesIO
import os
from pathlib import Path

__all__ = ['PDB']


#TODO Basically rewrite this whole thing
class PDB(Configurable):
    class Mode(Enum):
        AUTO = 0
        PDB = 1
        CIF = 2

    def __init__(self):
        self._current = BytesIO()
        self._mode = self.Mode.AUTO


    def SetCurrent(self, stream: BytesIO) -> None:
        self._current = stream

    def Get(self) -> BytesIO:
        return self._current

    def ToString(self) -> str:
        return self._current.read().decode()

    def clear(self) -> 'PDB':
        return self.Clear()

    def Clear(self) -> 'PDB':
        self._current = BytesIO()
        return self

    def auto(self):
        self._mode = self.Mode.AUTO

    def pdb(self):
        self._mode = self.Mode.PDB

    def cif(self):
        self._mode = self.Mode.CIF


    def UpdateSettings(self):
        pass

    # Enabling bad behavior
    def fetch(self, code: str, extension: str="mmCIF") -> 'PDB':
        return self.Fetch(code, extension=extension)

    # Fetches a file directly from the PDB
    def Fetch(self, code: str, extension: str="mmCIF") -> 'PDB':
        code = code.lower().strip()
        if "pdb" in extension.lower():
            filepath = f"pdb/{code[1:3]}/pdb{code}.ent.gz"
            name = f"pdb{code}.ent.gz"
        else:
            filepath = f"mmCIF/{code[1:3]}/{code}.cif.gz"
            name = f"{code}.cif.gz"
        fetch_url = f"https://files.rcsb.org/pub/pdb/data/structures/divided/{filepath}"
        # stream = cache.grab_url(fetch_url, name)
        stream = RemoteFileHandler.fetch_file_from_pdb(fetch_url, name) # TODO Why is this only a tar file on first download?
        if RemoteFileHandler.is_gzip(stream):
            stream = RemoteFileHandler.unpack_tar_gz_bio(stream)
        # print(stream.getvalue())
        self.SetCurrent(stream)
        return self

    def Load(self, file: str) -> 'PDB':
        if file.endswith(".gz"):
            with gzip.open(file, "r") as gz:
                stream = BytesIO(gz.read())
        else:
            with open(file, "rb") as fp:
                stream = BytesIO()
                stream.write(fp.read())
        self.SetCurrent(stream)
        return self


    #Currently only works for PDB files
    # Fetches a file from the PDB or cache and saves it to a location
    def SaveFetch(self, code: str, path: str, extension: str="mmCIF") -> bool:
        stream = self.Fetch(code, extension, save=False).Get()
        try:
            with open(path, "wb") as fp:
                fp.write(stream.read())
        except:
            return False

    def Construct(self, data: str|None = None):
        if os.path.exists(data):
            pass



    # TODO Everything below here :)


    # This function does too much.
    def Read(self, file: str|BytesIO):
        # Whatever we have, convert it to a list of string-lines
        if isinstance(file, str) and len(file) == 4: # We have a code
            lines = self.Fetch(file, extension='pdb', as_string=True).split('\n')
        elif isinstance(file, str): # We have a stringline
            pass
        elif isinstance(file, str) and os.path.exists(file): # Path to a file
            # Get file type
            # Read accordingly
            pass
        elif isinstance(file, BytesIO): # File stream
            pass
        else:
            raise ValueError("Data to read must be a PDB 4-letter code, a file as a string, a path to a file,"
                             "or a fetched file.")
        for line in lines:
            if line[0:6] == "ATOM  " or line[0:6] == "HETATM":
                print(line)
                atom = self._parse_atom_pdb(line)
                print(atom)
        # print(lines)

    def ParsePDB(self, lines):
        for line in lines:
            if line[0:6] == "ATOM  " or line[0:6] == "HETATM":
                print(line)
                atom = self._parse_atom_pdb(line)
                print(atom)

    def ParseCIF(self):
        pass

    # I guess people may want this?
    def ParseXML(self):
        pass

    def _prepare_file(self, file: str|BytesIO) -> list:
        if isinstance(file, str):
            file = file.strip()
        if isinstance(file, str) and len(file) == 4: # Try reading as a pdb code
            lines = self.Fetch(file, extension='pdb', as_string=True).split('\n')
        elif isinstance(file, str) and os.path.exists(file): # Path to a file
            lines = self.Load(file).split('\n')
        elif isinstance(file, str): # We have a stringline
            lines = file.split('\n')
        elif isinstance(file, BytesIO): # File stream
            lines = file.read().decode().split('\n')
        else:
            raise ValueError("Unable to process file.")
        return lines

    def _parse_atom_cif(self, line: str) -> Atom:
        pass

    def _parse_atom_pdb(self, line: str) -> Atom:
        record_name = line[0:6]
        serial = line[6:11]
        name = line[12:16]
        altLoc = line[16]
        resName = line[17:20]
        chainID = line[21]
        resSeq = line[22:26]
        iCode = line[26]
        x = line[30:38]
        y = line[38:46]
        z = line[46:54]
        occupancy = line[54:60]
        tempFactor = line[60:66]
        element = line[76:78]
        charge = line[78:80]
        return AtomFactory.EnforcedAtom(
            record_name=record_name,
            serial=serial,
            name=name,
            altLoc=altLoc,
            resName=resName,
            chainID=chainID,
            resSeq=resSeq,
            iCode=iCode,
            x=x,
            y=y,
            z=z,
            occupancy=occupancy,
            tempFactor=tempFactor,
            element=element,
            charge=charge
        )

    # Surprisingly hard to design this because of how irregular PDB files are
    # @classmethod
    # def DetectFiletype(cls, file: str|BytesIO, trust_extension: bool = True) -> list:
    #     if isinstance(file, str) and os.path.exists(file):
    #         if trust_extension and len(Path(file).suffixes) != 0:
    #             return Path(file).suffixes
    #     extensions = []
    #     lines = cls._prepare_file(file)
    #


    # @classmethod
    # def IsPDB(cls, file: str|BytesIO):
    #     pass
        # Extrapolate file type
        # Go line by line
        # Grab symmetry information
        # Parse atom information Organize residues


        # urllib.request.urlretrieve(fetch_url, name)
        # response = requests.get(fetch_url)
        # response.raise_for_status()
        # with gzip.GzipFile(fileobj=BytesIO(response.content)) as gz:
        #     with TextIOWrapper(gz) as fp:
        #         for line in fp.readlines():
        #             print(line.strip())
        #             break

        # print("Fetching ", prot)
        # import urllib.request
        # url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
        # try:
        #     with urllib.request.urlopen(url) as f:
        #         self.biomolecule_old.SetFetch(f.read().decode('utf-8'))
        #         self._Parse(hold_pdb)
        #         return True
        # except urllib.error.URLError:
        #     sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
        #     if crash:
        #         exit()
        #     return False

    # def Fetch(self, code: str, extension: str="pdb"):
    #
    #     fetch_url = r"https://files.rcsb.org/pub/pdb/data/structures/divided/pdb/b8/"
    #
    #     url = r"https://www.rcsb.org/fasta/entry/" + prot.strip().upper() + r"/display"
    #     try:
    #         with urllib.request.urlopen(url) as f:
    #             return f.read().decode('utf-8')
    #     except urllib.error.URLError:
    #         sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
    #         return False


# class PDB:
#     def __init__(self):
#         self.biomolecule_old = biomolecule_old()
#         # print(apalib.j_data.GetJson())
#
#     def Current(self):
#         return self.biomolecule_old
#
#     def FetchFASTA(self, prot):
#         import urllib.request
#         url = r"https://www.rcsb.org/fasta/entry/" + prot.strip().upper() + r"/display"
#         try:
#             with urllib.request.urlopen(url) as f:
#                 return f.read().decode('utf-8')
#         except urllib.error.URLError:
#             sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
#             return False
#
#     # def FetchAsFile(self, prot):
#     #     import urllib.request
#     #     url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
#     #     try:
#     #         with urllib.request.urlopen(url) as f:
#
#     #Enabling bad behavior
#     def fetch(self, prot, crash=True, hold_pdb=False):
#         return self.Fetch(prot, crash, hold_pdb)
#     def Fetch(self, prot, crash = True, hold_pdb=False):
#         # print("Fetching ", prot)
#         import urllib.request
#         url = r'https://files.rcsb.org/download/' + prot.strip() + '.pdb'
#         try:
#             with urllib.request.urlopen(url) as f:
#                 self.biomolecule_old.SetFetch(f.read().decode('utf-8'))
#                 self._Parse(hold_pdb)
#                 return True
#         except urllib.error.URLError:
#             sys.stderr.write("The requested pdb code could not be retrieved or does not exist\n")
#             if crash:
#                 exit()
#             return False
#
#     def Read(self, path, hold_pdb=False):
#         with open(path, 'r') as fp:
#             self.biomolecule_old.SetFetch(fp.read())
#             self._Parse(hold_pdb)
#
#     # Wrapper for the ParsePDB file to allow functionality with a fetched protein
#     def _Parse(self, hold_pdb=False):
#         try:
#             if self.biomolecule_old.GetFetch() is None:
#                 raise apaExcept.NoFetchError
#             return self._ParsePDB(self.biomolecule_old.GetFetch(), hold_pdb)
#             # return self._ParsePDB(self.container.GetFetch().splitlines())
#         except apaExcept.NoFetchError as e:
#             sys.stderr.write(e.message)
#
#
#     #PDB standard described here: https://www.wwpdb.org/documentation/file-format-content/format33/v3.3.html
#     def _ParsePDB(self, raw_pdb, hold_pdb=False):
#         self.biomolecule_old.ClearAll()
#         remark350 = ""
#         if hold_pdb:
#             self.biomolecule_old.SetFetch(raw_pdb)
#         for line in raw_pdb.splitlines():
#             # print(line)
#             if line[0:6] == 'ATOM  ' or line[0:6] == 'HETATM':
#                 self._ExtractAtomAndResidue(line)
#             if line.find("REMARK 350") != -1:
#                 remark350 += line + "\n"
#
#         symmetry_groups = self._ParseRemark350(remark350)
#         self.biomolecule_old._AddSymmetry(symmetry_groups)
#         self.biomolecule_old._PostParseEvaluations()
#     def _ParseRemark350(self, remark350):
#         lines = remark350.splitlines()
#         lines.append("END")
#         symFlag = False
#         biomolecules = []
#         for line in lines:
#             if 'REMARK 350 APPLY' in line and ":" in line and symFlag is False:
#                 symFlag = True
#                 symLines = []
#                 chains = line[line.find(":")+1:].replace(",", " ").split()
#             elif 'REMARK 350 APPLY' in line and ":" in line and symFlag is True:
#                 chains += line[line.find(":")+1:].replace(",", " ").split() #This feels dangerous
#             elif 'AND CHAINS:' in line and symFlag is True:
#                 chains += line[line.find(":") + 1:].replace(",", " ").split() #This also feels dangerous
#             elif 'BIOMT' in line and symFlag:
#                 BIOMT = line[13:19].strip()
#                 id = line[19:23].strip()
#                 x = line[23:33].strip()
#                 y = line[33:43].strip()
#                 z = line[43:53].strip()
#                 m = line[53:].strip()
#                 symLines.append([BIOMT, int(id), float(x), float(y), float(z) ,float(m)])
#             elif symFlag:
#                 symFlag = False
#                 biomolecule_old = {}
#                 biomolecule_old['chains'] = chains
#                 for sl in symLines:
#                     if sl[1] not in biomolecule_old.keys():
#                         biomolecule_old[sl[1]] = []
#                     biomolecule_old[sl[1]].append([sl[0]] + sl[2:])
#                 biomolecules.append(biomolecule_old)
#             #I hate PDB file format
#         for i in range(len(biomolecules)):
#             biomolecule_old = biomolecules[i]
#             for key in biomolecule_old.keys():
#                 biomolecule_old[key].sort(key=lambda x:x[0])
#                 for i in range(len(biomolecule_old[key])):
#                     if key == "chains":
#                         continue
#                     biomolecule_old[key][i].pop(0)
#         return biomolecules
#
#     def _ExtractAtomAndResidue(self, line):
#         serial = line[6:11].strip()
#         name = line[12:16].strip()
#         altLoc = line[16].strip()
#         resName = line[17:20].strip()
#         chainID = line[21].strip()
#         resSeq = line[22:26].strip()
#         iCode = line[26].strip()
#         x = line[30:38].strip()
#         y = line[38:46].strip()
#         z = line[46:54].strip()
#         occupancy = line[54:60].strip()
#         tempFactor = line[60:66].strip()
#         element = line[76:78].strip()
#         charge = line[78:80].strip()
#         atom = Atom.Atom(serial=serial, name=name, altLoc=altLoc, resName=resName, chainID=chainID, resSeq=resSeq,
#                          iCode=iCode, x=float(x), y=float(y), z=float(z), occupancy=occupancy, tempFactor=tempFactor, element=element,
#                          charge=charge)
#         if "HETATM" in line:
#             resType = "HETATM"
#         else:
#             resType = self.DetermineResType(resName)
#         residue = self.biomolecule_old.AddResidue(resType, resSeq, resName, chainID)
#         residue.InsertAtom(atom)
#
#     def DetermineResType(self, res_code):
#         if remode_data.ValidateRNA(res_code):
#             return 'RNA'
#         elif remode_data.ValidateDNA(res_code):
#             return 'DNA'
#         elif remode_data.ValidateAA(res_code):
#             return "AA"
#         else:
#             return "HETATM"
#
#     #Remove all of the waters from the current fetch. Probably make this more general for any HETATM group. Make a wrapper?
#     def RemoveWater(self):
#         h_chains = self.biomolecule_old.GetHETATMChains()
#         for chain in h_chains.keys():
#             h_chains[chain] = {key: value for (key, value) in h_chains[chain].items() if value.GetResName().upper() != 'HOH'}
#
#     # def Validate(self, **kwargs):
#     #     for key in kwargs:
#     #         if key != 'pdb' or (key == 'pdb' and not isinstance(kwargs['pdb'], str)):
#     #             raise apalib.apalibExceptions.BadKwarg('pdb=<pdb_to_validate>')
#
#     #Write contents to a PDB file
#
#     def AddChain(self, collection, name=None):
#         if name is None:
#             chains = list(self.biomolecule_old.Chains.keys())
#             for i in range(26):
#                 if chr(ord('z') - i) not in chains:
#                     name = chr(ord('z') - i)
#                     break
#         else:
#             if len(name) > 1 or not name.isalnum():
#                 raise BadNameException("A chain must be a single-character alphanumeric input")
#         #Rechain all of the atoms and residues with new name
#         for residue in collection:
#             residue.SetChainID(name)
#             for atom in residue.GetAtoms():
#                 atom.SetChainID(name)
#         self.biomolecule_old.AddChain(name)
#         self.biomolecule_old.Chains[name] = collection
#
#     def WritePDB(self, fp):
#         s = sorted(self.biomolecule_old.DumpResidues(), key=lambda x: x.seqNum)
#         with open(fp, "w") as f:
#             for res in s:
#                 f.write(res.WriteForPDB())
#
#     #Write contents to FASTA
#     def ToFASTA(self):
#         ls = self.biomolecule_old.AsList(ordered=True)
#         retStr = ""
#         for r in ls:
#             if remode_data.ValidateAA(r.resName):
#                 name = remode_data.Map("Amino Acids", r.resName)
#                 retStr += remode_data.GetJson()["Amino Acids"][name]["1code"]
#             elif r.resName.upper() != "HOH":
#                 retStr += "X"
#         return retStr


from io import BytesIO
import requests
import gzip
from urllib.request import urlopen
import aldepyde


class RemoteFileHandler():
    @staticmethod
    def fetch_file_from_pdb(url: str, name) -> BytesIO:
        cache = aldepyde.get_cache()
        if cache.in_cache(name):
            return cache.extract_from_cache(name)
        # Return a requested file as a BytesIO stream from a URL or the cache
        response = requests.get(url)
        response.raise_for_status()
        stream_io = BytesIO(response.content)
        aldepyde.get_cache().save_to_cache(stream_io, name)
        return stream_io

    @staticmethod
    def fetch_file(url: str) -> BytesIO:
        response = urlopen(url)
        return BytesIO(response.read())

    @staticmethod
    def is_gzip(stream: BytesIO) -> bool:
        magic_start = stream.read(2)
        stream.seek(0)
        return magic_start == b'\x1f\x8b'

    @staticmethod
    def unpack_tar_gz_bio(stream: BytesIO) -> BytesIO:
        with gzip.open(stream, "r") as gz:
            return BytesIO(gz.read())





from ._amino_acid import *
from ._Atom import *
from ._AtomFactory import *
from ._dna import *
from ._pdb import *
# from .Residue import * # This probably doesn't need to be revealed to the user
from ._rna import *

__all__ =  list(set(_amino_acid.__all__.copy()) |
           set(_Atom.__all__.copy()) |
           set(_AtomFactory.__all__.copy()) |
           set(_dna.__all__.copy()) |
           set(_pdb.__all__.copy()) |
           set(_rna.__all__.copy()))

import sys

sys.stderr.write("Note that the `biomolecule_old` submodule is not yet fully tested and may be unstable")


from typing import Literal
from aldepyde.data import get_store


def hydropathy(sequence: str, scale: Literal['kd'] = 'kd', window: int = 0, window_average: bool = False) -> list:
    hydropathies = []
    den = len(sequence) if window_average\
    else 1
    if window < 0:
        return [sum([get_store().hydropathy(aa, scale=scale) for aa in sequence])/den]
    for i in range(window, len(sequence) - window, 1):
        residue_window = sequence[i - window:i + window + 1]
        hydropathies.append(sum([get_store().hydropathy(aa, scale=scale) for aa in residue_window])/(len(residue_window) if window_average
                                                                                                    else 1))
    if len(hydropathies) == 0:
        return [sum([get_store().hydropathy(aa, scale=scale) for aa in sequence])/den]
    return hydropathies

def helical_prop(sequence: str, window: int = 0, window_average: bool = False) -> list:
    propensities = []
    den = len(sequence) if window_average\
    else 1
    if window < 0:
        return [sum([get_store().helical_prop(aa) for aa in sequence])/den]
    for i in range(window, len(sequence) - window, 1):
        residue_window = sequence[i - window:i + window + 1]
        propensities.append(sum([get_store().helical_prop(aa) for aa in residue_window])/(len(residue_window) if window_average
                                                                                        else 1))
    if len(propensities) == 0:
        return [sum([get_store().helical_prop(aa) for aa in sequence])/den]
    return propensities

# TODO Implement this algorithm
# https://bio.libretexts.org/Courses/Ouachita_Baptist_University/Reyna_Cell_Biology/02%3A_2-(T2-first_lecture)_Protein_Structure/2.05%3A_G._Predicting_Protein_Properties_From_Sequences/2.5.02%3A_G2._Prediction_of_Secondary_Structure

def chou_fasman():
    pass


def smith_waterman(sequence_1:str, sequence_2:str, match:int=1, gap:int=-2, mismatch:int=-1, minimum=0, recurse:bool=True) -> tuple[float, tuple[str, str]]:
    matrix = [[0] * (len(sequence_1) + 1)]
    pointer_matrix = [[[(-1,-1)]] * (len(sequence_1) + 1)]
    max_value = -float('inf')
    max_coords = list()
    coords = (-1, -1)
    for i in range(1, len(sequence_2) + 1):
        matrix.append(list([0]))
        pointer_matrix.append(list([[(-1, -1)]]))
        for v in range(1, len(sequence_1) + 1):
            u = (matrix[i-1][v] + gap, 'u')
            l = (matrix[i][v-1] + gap, 'l')
            d = (matrix[i-1][v-1] + match if sequence_1[v-1] == sequence_2[i-1] else matrix[i-1][v-1] + mismatch, 'd')
            s = (0, 's')
            scores = [u, l, d, s]
            max_sub_score = max([x[0] for x in scores])
            directions = []
            val = -1
            for value_dir in scores:
                if value_dir[0] == max_sub_score:
                    val = max_sub_score
                    if value_dir[1] == 'd':
                        directions.append((i-1, v-1))
                    elif value_dir[1] == 'u':
                        directions.append((i-1,v))
                    elif value_dir[1] == 'l':
                        directions.append((i,v-1))
            pointer_matrix[-1].append(directions)
            matrix[-1].append(val)
            if val > max_value:
                max_value = val
                max_coords = list()
            if val == max_value:
                max_coords.append((i,v))

    #traceback
    # TODO Make recursion optional
    import copy
    results = []
    def pointer_helper(al1, al2, mc):
        # print(pointer_matrix[mc[0]][mc[1]])
        curr_1 = al1
        curr_2 = al2
        for new_point in pointer_matrix[mc[0]][mc[1]]:
            c1 = sequence_1[mc[1] - 1] if mc[1] - 1 == new_point[1] else '-'
            c2 = sequence_2[mc[0] - 1] if mc[0] - 1 == new_point[0] else '-'

            al1.append(c1)
            al2.append(c2)
            # al1.append(sequence_1[mc[1] - 1]) if mc[1] - 1 == new_point[1] else al1.append('-')
            # al2.append(sequence_2[mc[0] - 1]) if mc[0] - 1 == new_point[0] else al2.append('-')
            if matrix[new_point[0]][new_point[1]] != 0:
                pointer_helper(al1, al2, new_point)
                # al1, al2 = pointer_helper(copy.deepcopy(curr_1), copy.deepcopy(curr_2), new_point)
            else:
                results.append(("".join(reversed(al1)), "".join(reversed(al2))))
            al1.pop()
            al2.pop()
        # return al1, al2

    alignments = list()
    for mc in max_coords:
        pointer_helper(list(), list(), mc)
    for result in results:
        alignments.append((result[0], result[1]))
    print(alignments)
    return max_value, alignments



import os
import re

def _verify_cache_directory(path: str) -> bool:
    return os.path.exists(os.path.join(path))

def _parse_memory(memory: str) -> int:
    ALLOWED_PREFIX = "bkmgt" # No, you don't get to use petabytes
    full_re = f"[0-9]+[{ALLOWED_PREFIX}]?i?b?"
    numeric_re = "[0-9]+"

    if memory.isnumeric():
        memory += "mib"
    if re.fullmatch(full_re, memory, flags=re.IGNORECASE) is None:
        raise ValueError(f"Requested memory must be of the following form: {full_re}")

    match = re.match(numeric_re, memory)
    numeric = int(memory[:match.span()[1]])
    unit = memory[match.span()[1]:]
    base = 1024 if "i" in unit else 1000
    multiple = base**(ALLOWED_PREFIX.index(unit[0].lower()))
    return numeric * multiple

def _convert_memory_bits(memory: int) -> str:
    ALLOWED_PREFIX = "bkmgt"
    digits = len(str(memory))
    return f"{memory / (1000 ** (digits//3)):.3f} {ALLOWED_PREFIX[digits//3].upper()}b"

def _convert_memory_bytes(memory: int) -> str:
    ALLOWED_PREFIX = "bkmgt"
    digits = len(str(memory))
    return f"{memory / (1024 ** (digits // 3)):.3f} {ALLOWED_PREFIX[digits // 3].upper()}b"


from io import BytesIO
from dataclasses import dataclass, field
from typing import List, Tuple, Optional, Dict
from .pdb_client import pdb_Client
import gzip
from .parse_mmcif import _parse_mmcif
from .parse_pdb import _parse_pdb
from .parse import parse
from .load import load
from .fetch import fetch_text, fetch_raw, fetch

__all__ = ["fetch", "fetch_raw", "fetch_text", "load", "parse"]


import math
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass


@dataclass
class ConstraintSpecs():
    categories:dict[str,str]
    constraints:dict[str,float]=None

    def __post_init__(self):
        if self.constraints is None:
            self.set_constraints_from_categories()

    def set_constraints_from_categories(self):
        self.constraints = {k:1 for k in self.categories}

    def normalize(self, establish=True):
        non_zero = [x for x,v in self.categories.items() if len(v) > 0]
        total = sum([self.constraints[x] for x in non_zero])
        normalized = {k:(v/total) if len(self.categories.get(k)) > 0 else 0 for k, v in self.constraints.items() }
        if establish:
            self.constraints = normalized
        return normalized

    def clear_all(self):
        self.categories = dict()
        self.constraints = dict()

    def clear_category(self, category:str):
        self.categories[category] = ""

    def clear_constraint(self, category:str):
        self.constraints[category] = 0

    def clear_constraints(self):
        self.constraints = {k:0 for k in self.constraints}

    def clear_alphabet(self):
        self.categories = {k:"" for k in self.categories}

    def wipe(self, establish=False) -> dict:
        c =  {k:1 for k in self.constraints}
        if establish:
            self.constraints = c
        return self.constraints

    def set_category_constraint(self, category:str, constraint:float):
        self.constraints[category] = constraint

    def hard_set_category_constraint(self, category:str, percentage:float):
        if percentage > 1:
            raise ValueError("Percentage must be less than 1.0")
        self.constraints[category] = 0
        self.normalize()
        self.constraints = {k:v*(1-percentage) for k,v in self.constraints.items()}
        self.constraints[category] = percentage


    #TODO Consider what you want to do with wiping/not wiping
    def combine_categories(self, category_1: str, category_2: str, new_category=None, constraint_ratio: float = 1.0):
        if new_category is None:
            new_category = category_1
        s1 = self.categories.pop(category_1)
        s2 = self.categories.pop(category_2)
        self.constraints.pop(category_1)
        self.constraints.pop(category_2)
        self.constraints[new_category] = constraint_ratio
        self.categories[new_category] = frozenset().union(s1, s2)
        if constraint_ratio == -1:
            self.wipe()

    def get_ratio(self, category: str) -> float:
        return self.constraints.get(category)

    def get_constraints(self) -> dict:
        return self.constraints

    def get_category(self, category:str):
        return self.categories.get(category)

    def get_categories(self) -> dict:
        return self.categories

    def get_expected(self, length:int):
        constraint_ratios = self.normalize()
        pairs = sorted([[k, length * v] for (k, v) in constraint_ratios.items()], key=lambda x: x[1] % 1, reverse=True)
        floored = [[x[0], math.floor(x[1])] for x in pairs]
        total_residues = sum([x[1] for x in floored])
        index = 0
        while total_residues < length:
            floored[index][1] += 1
            total_residues += 1
            index += 1
        return {k: math.floor(v) for (k, v) in floored}

    def _is_in(self, value) -> str:
        for key in self.categories:
            if value in self.categories.get(key):
                return key
        return '*'

    def is_expected(self, sequence:str, min_tolerance:int=1) -> bool:
        expected = self.get_expected(length=len(sequence))
        actual_per_category = {k:0 for k in expected}
        actual_count = Counter(sequence)
        for key in actual_count:
            actual_per_category[self._is_in(key)] += actual_count[key]
        normalized = self.normalize(establish=False)
        for key in actual_per_category:
            p_or_m = max(min_tolerance, math.floor(math.sqrt(len(sequence)*(normalized[key]))))
            if not expected[key] - p_or_m <= actual_per_category[key] <= expected[key] + p_or_m:
                return False
        return True



    @property
    def alphabet(self) -> set:
        return set().union(*[val for val in self.categories.values()])


    #TODO Implement the following methods:
    # get_ratio()
    # get_ratios()
    # get_category()
    # get_categories()
    # get_expected()
    # is_expected()


# Define constraints when generating a polymer
class PolymerConstraints(ABC):
    def __init__(self, specs:ConstraintSpecs):
        self.sps = specs

    # @abstractmethod
    # def get_default_classes(self):
    #     pass

    def get_specs(self) -> ConstraintSpecs:
        return self.sps

    @property
    def specs(self) -> ConstraintSpecs:
        return self.sps

    # def new_constraint_ratios(self) -> dict:
    #     # return self.normalize_table({k:1 for k in self.classes})
    #     return self.specs.normalize(self.specs.wipe(establish=False))

    # def set_constraint_classes(self, table:dict):
    #     self.classes = table
    #
    # def normalize_table(self, table:dict) -> dict:
    #     non_zero = [x for x,v in self.classes.items() if len(v) > 0]
    #     total = sum([table[x] for x in non_zero])
    #     return {k:(v/total) if len(self.classes.get(k)) > 0 else 0 for k, v in table.items() }
    #
    # def normalize(self):
    #     self.constraint_ratios = self.normalize_table(self.constraint_ratios)
    #
    # def set_constraint_ratios(self, table:dict):
    #     self.constraint_ratios = self.normalize_table(table)
    #
    # def set_class_ratio(self, class_name:str, value:float, normalize:bool=False):
    #     self.constraint_ratios[class_name] = value
    #     if normalize:
    #         self.constraint_ratios = self.normalize_table(self.constraint_ratios)
    #
    # def hard_set(self, class_name:str, percentage:float):
    #     if percentage > 1:
    #         raise ValueError("Percentage must be less than 1.0")
    #     self.constraint_ratios[class_name] = 0
    #     self.normalize()
    #     self.constraint_ratios = {k:v*(1-percentage) for k,v in self.constraint_ratios.items()}
    #     self.constraint_ratios[class_name] = percentage
    #
    # def combine(self, class_1:str, class_2:str, new_class:str, ratio:float=-1):
    #     s1 = self.classes.pop(class_1)
    #     s2 = self.classes.pop(class_2)
    #     self.constraint_ratios.pop(class_1)
    #     self.constraint_ratios.pop(class_2)
    #     self.constraint_ratios[new_class] = ratio
    #     self.classes[new_class] = frozenset().union(s1, s2)
    #     if ratio == -1:
    #         self.wipe()
    #
    # def alphabet(self) -> set:
    #     return set().union(*[val for val in self.classes.values()])
    #
    # #TODO Get this working
    #
    # # def remove(self, name:str):
    # #     self.classes.pop(name)
    # #     self.constraint_ratios = self.normalize_table(self.constraint_ratios)
    #
    # def negate(self, name:str):
    #     self.constraint_ratios[name] = 0
    #     self.constraint_ratios = self.normalize_table(self.constraint_ratios)
    #
    # def wipe(self):
    #     self.constraint_ratios = {k:1 for k in self.constraint_ratios}
    #
    # def clear_all(self):
    #     self.constraint_ratios = {k:0 for k in self.constraint_ratios}
    #
    # def expected_per_class(self, length):
    #     constraint_ratios = self.normalize_table(self.constraint_ratios)
    #     pairs = sorted([[k,length*v] for (k, v) in constraint_ratios.items()], key=lambda x: x[1]%1, reverse=True)
    #     floored = [[x[0], math.floor(x[1])] for x in pairs]
    #     total_residues = sum([x[1] for x in floored])
    #     index = 0
    #     while total_residues < length:
    #         floored[index][1] += 1
    #         total_residues += 1
    #         index += 1
    #     return {k:math.floor(v) for (k, v) in floored}



from .residue_generator import ResidueGenerator
from .protein_specs import ProteinClassifier
from .protein_constraints import ProteinConstraints

class ProteinGenerator(ResidueGenerator):
    def __init__(self, classifier:ProteinClassifier|None = None):
        if classifier is None:
            super().__init__(ProteinClassifier())
        else:
            super().__init__(classifier)

    def get_classifier(self):
        return self.classifier

    def generate(self, length:int, k:int=1, constraints:ProteinConstraints=ProteinConstraints()):
        return super().generate(length=length, k=k, constraints=constraints)




from .sniff import _detect_gz
from .parse import parse
import gzip


def load(filepath):
    if _detect_gz(filepath):
        with gzip.open(filepath, 'rb') as gz:
            text = gz.read().decode('utf-8')
    else:
        with open(filepath, 'r') as fp:
            text = fp.read()
    return parse(text)



from typing import Literal, Tuple

from aldepyde.data import get_store, DataStore
from aldepyde.biomolecule import Residue, Atom

# Determine if a residue is DNA, RNA, or an amino acid
# 1. Check atom list for backbone
# 2. Check sidechain for necessary atoms
# 3. Parse name
def residue_type(residue:Residue, is_het:bool|None) -> Literal['amino_acid', 'dna', 'rna', 'hetatm']|None:
    if is_het:
        return 'hetatm'
    res_atoms = residue.dump_atom_labels(lower=True)
    telling_nucleotide = ["c1'", "c2'", "c3'", "c4'", "c5'", "o3'", "o5'", "p", "op1", "op2"]
    telling_rna = ["o2'"]
    telling_aa = ["c", "n", "o"]

    if len([x for x in res_atoms if x in telling_rna]) > 0:
        return 'rna'
    elif len([x for x in res_atoms if x in telling_aa]) > 0:
        return 'amino_acid'



    # This is technically robust by PDB/mmCIF standards
    if len(residue.name) == 3:
        return 'amino_acid'
    elif len(residue.name) == 2:
        return 'dna'
    elif len(residue.name) == 1:
        return 'rna'

def centroid(residue:Residue, res_type) -> Tuple[float, float, float] | None:
    centroid_atoms = get_store().centroid_atoms(residue.name, res_type=res_type)
    included_atoms = residue.get_atoms_by_label(centroid_atoms)
    try:
        x = sum([a.x for a in included_atoms]) / len(included_atoms)
        y = sum([a.y for a in included_atoms]) / len(included_atoms)
        z = sum([a.z for a in included_atoms]) / len(included_atoms)
        return (x, y, z)
    except ZeroDivisionError:
        return None

# TODO This method will surprisingly be a massive pain. For now only use if safely pulled from a file or provided
# because wouldn't you know... there's no PDB standard apparently!
def derive_element(atom:Atom) -> str|None:
    return atom.element.title()

def get_element_mass(symbol:str) -> float:
    return get_store().atomic_mass(symbol)


def center_of_mass(residue:Residue, res_type) -> Tuple[float, float, float] | None:
    centroid_atoms = get_store().centroid_atoms(residue.name, res_type=res_type)
    included_atoms = residue.get_atoms_by_label(centroid_atoms)
    point_tups = []
    for atom in included_atoms:
        point_tups.append((atom, get_element_mass(derive_element(atom))))
    try:
        x = sum([a[0].x * a[1] for a in point_tups]) / sum([x[1] for x in point_tups])
        y = sum([a[0].y * a[1] for a in point_tups]) / sum([x[1] for x in point_tups])
        z = sum([a[0].z * a[1] for a in point_tups]) / sum([x[1] for x in point_tups])
        return (x, y, z)
    except ZeroDivisionError:
        return None




import textwrap

def to_fasta(*args: tuple[str,str,str], output:str|None = None, max_length:int = 80) -> None:
    if output is None:
        raise ValueError("Output may not be `None`. Please specify an output location")
    with open(output, 'w') as fp:
        for tup in args:
            name, desc, sequence = tup
            fp.write(f'>{name}|{desc}\n')
            for chunk in textwrap.wrap(sequence, width=max_length):
                fp.write(f"{chunk}\n")


test = "MSQIIWGAYAQRNTEDHPPAYAPGYKTSVLRSPKNALISIAETLSEVTAPHFSADKFGPKDNDLILNYAKDGLPIGERVIVHGYVRDQFGRPVKNALVEVWQANASGRYRHPNDQYIGAMDPNFGGCGRMLTDDNGYYVFRTIKPGPYPWRNRINEWRPAHIHFSLIADGWAQRLISQFYFEGDTLIDSCPILKTIPSEQQRRALIALEDKSNFIEADSRCYRFDITLRGRRATYFENDLT"
test2 = "MSQIIWGAYAQRNTEDHPPAYAPGYKTSVLRSPKNALISIAETLSEVTAPHFSADKFGPKDNDLILNYAKDGLPIGERVIVHGYVRDQFGRPVKNALVEVWQANASGRYRHPNDQYIGAMDPNFGGCGRMLTDDNGYYVFRTIKPGPYPWRNRINEWRPAHIHFSLIADGWAQRLISQFYFEGDTLIDSCPILKTIPSEQQRRALIALEDKSNFIEADSRCYRFDITLRGRRATYFENDLT"

to_fasta(('test', 'test sequence', test), ('test', 'test sequence', test2), output=r"C:\Users\mcmurran\Downloads\test.fasta")


from io import BytesIO, TextIOWrapper
import requests
import gzip


def _fetch_file_from_pdb(url: str, filename: str) -> BytesIO:
    # Return a requested file as a BytesIO stream from a URL or the cache
    response = requests.get(url)
    response.raise_for_status()
    stream_io = BytesIO(response.content)
    return stream_io





from abc import ABC, abstractmethod
import gzip
import urllib.request
import urllib.error
# import requests
import os
from typing import Tuple, BinaryIO, Optional
from io import TextIOWrapper

class streamable_database(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def fetch(self, url):
        pass

    @abstractmethod
    def fetch_code(self, codes):
        pass

    @abstractmethod
    def parse(self, text):
        pass

    @staticmethod
    def open_stream(source:str, provide_headers=False) -> Tuple[BinaryIO, int] | Tuple[BinaryIO, int,  dict] | None:
        if source.startswith('http://') or source.startswith('https://'):
            req = urllib.request.Request(source, headers={"User-Agent": "aldepyde/streamer"})
            try:
                resp = urllib.request.urlopen(req)
                length = resp.headers.get("Content-Length")
                if provide_headers:
                    return resp, int(length) if length else None, resp.headers
                else:
                    return resp, int(length) if length else None
            except urllib.error.HTTPError as e:
                raise
            except urllib.error.URLError as e:
                raise
            # resp = requests.get(source, stream=True)
            # resp.raise_for_status()
            # length = resp.headers.get("Content-Length")
            # return resp.raw, int(length) if length else None
        else:
            size = os.path.getsize(source)
            return open(source, 'rb'), size



    # Yes, I know the first conditionals do the same thing

    def __call__(self):
        pass

class local_database(ABC):

    def __init__(self, filepath=None, as_fp=False):
        self.fp = None
        self.as_fp = as_fp
        self.size = None
        self.load_path(filepath)

    def refresh(self):
        if self.fp is not None:
            self.fp.seek(0)

    def load_path(self, filepath):
        self.filepath = filepath

    def get_pointer(self):
        return self.fp

    def __enter__(self):
        self.fp, self.size = local_database.open_stream(self.filepath)
        if self.as_fp:
            return self.fp
        else:
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fp is not None:
            self.fp.close()
        self.fp = None

    @staticmethod
    def open_stream(source:str) -> Tuple[BinaryIO, int] | None:
        size = os.path.getsize(source)
        return open(source, 'rb'), size


import zlib
import re
from aldepyde.databases._database import streamable_database
from aldepyde.distribution.distributions.standards import source_head, distribution_head
from aldepyde.utils import ProgressBar
import os
import json
from dataclasses import dataclass


@dataclass
class UniProtFastaRecord():
    header: str
    sequence: str

class uniref_parser(streamable_database):
    def __init__(self):
        super().__init__()

    # TODO single entry parsing
    # TODO store metadata upon request
    # TODO implement abstract methods



    @staticmethod
    def _stream_uniref_plain(path, chunk_size=8192, use_progress_bar=False, compressed=False):
        raw_stream, size = streamable_database.open_stream(path)
        pbar = ProgressBar(size//chunk_size) if use_progress_bar else None
        decompressor = zlib.decompressobj(16 + zlib.MAX_WBITS)
        try:
            while True:
                if not compressed:
                    decomp_chunk = raw_stream.read(chunk_size)
                else:
                    comp_chunk = raw_stream.read(chunk_size)
                    decomp_chunk = decompressor.decompress(comp_chunk)
                    if comp_chunk == b"":
                        break
                if not decomp_chunk:
                    break
                if pbar is not None:
                    pbar.update()
                yield decomp_chunk
            final = decompressor.flush()
            if final:
                yield final
        finally:
            raw_stream.close()

    @staticmethod
    def _next_link(headers) -> str|None:
        link_statement = headers.get('Link')
        if link_statement is None:
            return None
        m = re.match(r'<([^>]+)>', link_statement)
        return m.group(1)

    @staticmethod
    def _stream_paginated_plain(url, chunk_size=8192, compressed=False):
        try:
            while True:
                raw_stream, size, headers = streamable_database.open_stream(url, provide_headers=True)
                decompressor = zlib.decompressobj(16 + zlib.MAX_WBITS)

                while True:
                    if not compressed:
                        decomp_chunk = raw_stream.read(chunk_size)
                    else:
                        comp_chunk = raw_stream.read(chunk_size)
                        decomp_chunk = decompressor.decompress(comp_chunk)
                    if not decomp_chunk:
                        break
                    yield decomp_chunk
                final = decompressor.flush()
                if final:
                    yield final
                url = uniref_parser._next_link(headers)
                # print(url)
                if url is None:
                    break
                raw_stream.close()
        finally:
            raw_stream.close()

    @staticmethod
    def stream_uniref_gz(filepath, chunk_size=8192, use_progress_bar=False):
        raw_stream, size = streamable_database.open_stream(filepath)
        pbar = ProgressBar(size//chunk_size) if use_progress_bar else None
        decompressor = zlib.decompressobj(16 + zlib.MAX_WBITS)
        try:
            while True:
                comp_chunk = raw_stream.read(chunk_size)
                if not comp_chunk:
                    break
                if pbar is not None:
                    pbar.update()
                decomp_chunk = decompressor.decompress(comp_chunk)
                if decomp_chunk:
                    yield decomp_chunk
            final = decompressor.flush()
            if final:
                yield final
        finally:
            raw_stream.close()

    @staticmethod
    def download_paginated(url, destination, chunk_size=8192):
        with open(destination, 'wb') as fp:
            while True:
                raw_stream, size, headers = streamable_database.open_stream(url, provide_headers=True)
                while True:
                    chunk = raw_stream.read(chunk_size)
                    if not chunk:
                        break
                    fp.write(chunk)
                url = uniref_parser._next_link(headers)
                if url is None:
                    break
                raw_stream.close()

    @staticmethod
    def download_file(url, destination, chunk_size=8192, use_progress_bar=False):
        raw_stream, size = streamable_database.open_stream(url)
        pbar = ProgressBar(size // chunk_size) if use_progress_bar else None
        with open(destination, 'wb') as fp:
            while True:
                chunk = raw_stream.read(chunk_size)
                if not chunk:
                    break
                if pbar is not None:
                    pbar.update()
                fp.write(chunk)


    @staticmethod
    def _stitch_streamed_sequences(stream):
        buffer = b''
        one_record_re = re.compile(br'\A>.*(?:\n(?!>).*?)*(?=\n>|$)')
        for chunk in stream:
            buffer += chunk
            # Take out just one entry at a time. Don't be a hero and shatter the whole chunk
            while buffer.count(b'\n>') > 0:
                m = one_record_re.match(buffer)
                entry = m.group(0)
                buffer = buffer[len(entry) + 1:]
                spl = entry.split(b'\n')
                header = spl[0].decode('utf-8')
                sequence = b''.join(spl[1:]).decode('utf-8')
                yield UniProtFastaRecord(header=header, sequence=sequence)
        spl = buffer.split(b'\n')
        header = spl[0].decode('utf-8')
        sequence = b''.join(spl[1:]).decode('utf-8')
        yield UniProtFastaRecord(header=header, sequence=sequence)
        #         yield  b"".join(entry.split(b'\n')[1:])
        # yield b"".join(buffer.split(b'\n')[1:])

        #         sequences = [b">" + seq for seq in buffer.split(b">") if seq != b""]
        #         buffer = buffer[buffer.rfind(b">"):]
        #         ret_l = [b"".join(sequence.split(b'\n')[1:]).replace(b"\n", b"") for sequence in sequences[:-1]]
        #         for s in ret_l:
        #             yield s if not as_str else s.decode()
        # yield uniref_parser._final_sequence(buffer) if not as_str else uniref_parser._final_sequence(buffer).decode()

    @staticmethod
    def _final_sequence(buffer):
        lines = buffer.split(b'\n')
        return b"".join(lines[1:])

    @staticmethod
    def stream_file(filepath, chunk_size=8192, use_progress_bar=False, stitch=False):
        with open(filepath, 'rb') as fp:
            magic_start = fp.read(2)
        compressed = True if magic_start == b'\x1f\x8b' else False
        if not stitch:
            yield from uniref_parser._stream_uniref_plain(filepath, chunk_size=chunk_size,
                                                      use_progress_bar=use_progress_bar, compressed=compressed)
        else:
            yield from uniref_parser._stitch_streamed_sequences(uniref_parser._stream_uniref_plain(filepath, chunk_size=chunk_size,
                                                      use_progress_bar=use_progress_bar, compressed=compressed))

    @staticmethod
    def stream_query(url, chunk_size=8192, use_progress_bar=False, stitch=False):
        paginated = True if 'size' in url else False
        compressed = True if 'compressed' in url else False
        if not paginated:
            if not stitch:
                yield from uniref_parser._stream_uniref_plain(url, chunk_size=chunk_size, use_progress_bar=use_progress_bar, compressed=compressed)
            else:
                yield from uniref_parser._stitch_streamed_sequences(uniref_parser._stream_uniref_plain(
                    url,chunk_size=chunk_size, use_progress_bar=use_progress_bar, compressed=compressed))
        else:
            if not stitch:
                yield from uniref_parser._stream_paginated_plain(url, chunk_size=chunk_size, compressed=compressed)
            else:
                yield from uniref_parser._stitch_streamed_sequences(uniref_parser._stream_paginated_plain(
                    url,chunk_size=chunk_size, compressed=compressed))

    @staticmethod
    def download_query(url, destination, chunk_size=8192, use_progress_bar=False):
        paginated = True if 'size' in url else False
        if not paginated:
            uniref_parser.download_file(url, destination=destination,
                                        chunk_size=chunk_size, use_progress_bar=use_progress_bar)
        else:
            uniref_parser.download_paginated(url, destination=destination,
                                        chunk_size=chunk_size)

    @staticmethod
    def stream_uniref50(chunk_size=8192, use_progress_bar=False, stitch=False):
        if not stitch:
            yield from uniref_parser.stream_uniref_gz('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref50/uniref50.fasta.gz',
                                                  chunk_size=chunk_size, use_progress_bar=use_progress_bar)
        else:
            yield from uniref_parser._stitch_streamed_sequences(uniref_parser.stream_uniref_gz(
                'https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref50/uniref50.fasta.gz',
                                                  chunk_size=chunk_size, use_progress_bar=use_progress_bar))

    @staticmethod
    def stream_uniref90(chunk_size=8192, use_progress_bar=False, stitch=False):
        if not stitch:
            yield from uniref_parser.stream_uniref_gz('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref90/uniref90.fasta.gz',
                                                  chunk_size=chunk_size, use_progress_bar=use_progress_bar)
        else:
            yield from uniref_parser._stitch_streamed_sequences(uniref_parser.stream_uniref_gz(
                'https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref90/uniref90.fasta.gz',
                chunk_size=chunk_size, use_progress_bar=use_progress_bar))

    @staticmethod
    def stream_uniref100(chunk_size=8192, use_progress_bar=False, stitch=False):
        if not stitch:
            yield from uniref_parser.stream_uniref_gz('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref100/uniref100.fasta.gz',
                                                  chunk_size=chunk_size, use_progress_bar=use_progress_bar)
        else:
            yield from uniref_parser._stitch_streamed_sequences(uniref_parser.stream_uniref_gz(
                'https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref100/uniref100.fasta.gz',
                chunk_size=chunk_size, use_progress_bar=use_progress_bar))

    @staticmethod
    def download_uniref50(destination='uniref50.fasta.gz', chunk_size=8192, use_progress_bar=False):
        uniref_parser.download_file('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref50/uniref50.fasta.gz', destination=destination,
                                    chunk_size=chunk_size, use_progress_bar=use_progress_bar)

    @staticmethod
    def download_uniref90(destination='uniref90.fasta.gz', chunk_size=8192, use_progress_bar=False):
        uniref_parser.download_file('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref90/uniref90.fasta.gz', destination=destination,
                                    chunk_size=chunk_size, use_progress_bar=use_progress_bar)
    @staticmethod
    def download_uniref100(destination='uniref100.fasta.gz', chunk_size=8192, use_progress_bar=False):
        uniref_parser.download_file('https://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref100/uniref100.fasta.gz', destination=destination,
                                    chunk_size=chunk_size, use_progress_bar=use_progress_bar)



            # yield from uniref_parser._stream_uniref_plain(filepath, chunk_size=chunk_size,
            #                                               use_progress_bar=use_progress_bar, stitch=stitch)

    @staticmethod
    def aa_frequency(path:str, json_dest:str|None = None, as_probs=False, chunk_size=8192, use_progress_bar=False) -> dict:
        is_filepath = True if os.path.isfile(path) else False
        if is_filepath:
            generator = uniref_parser.stream_file(path, chunk_size=chunk_size, use_progress_bar=use_progress_bar, stitch=True)
        else:
            generator = uniref_parser.stream_query(path, chunk_size=chunk_size, use_progress_bar=use_progress_bar, stitch=True)
        occurances = dict()
        for record in generator:
            for c in record.sequence:
                if c not in occurances:
                    occurances[c] = 0
                occurances[c] += 1
        if as_probs:
            total = sum(occurances.values())
            for key in occurances:
                occurances[key] = occurances[key] / total
        if json_dest is not None:
            json_dict = {
                distribution_head : occurances,
                source_head : path
            }
            with open(json_dest, 'w') as fp:
                json.dump(json_dict, fp, indent=3)
        return occurances


from .polymer_specs import PolymerClassifier
from .defaults.protein_defaults import ProteinDefaults
from dataclasses import dataclass, field
from typing import Mapping

class ProteinClassifier(PolymerClassifier):
    def __init__(self, alphabet: str | list | set = None):
        ab = self.get_default_alphabet() if alphabet is None else alphabet
        # cs = self.get_default_classes() if classes is None else classes
        # fl = self.default_flags() if flags is None else flags
        super().__init__(ab)

    def get_default_alphabet(self):
        return ProteinDefaults().get_alphabet()

    def get_default_classes(self):
        return ProteinDefaults().get_classes()

    def set_default_alphabet(self):
        self.alphabet = ProteinDefaults().get_alphabet()

    def set_default_classes(self):
        self.classes = ProteinDefaults().get_classes()



    # def default_flags(self):
    #     self.flags =ProteinDefaults.flags







import zlib
from io import BytesIO
import urllib.request
import gzip
from aldepyde import get_cache


GZIP = b"\x1f\x8b"
ZIP = b"\x50\x4B\x03\x04"

class RemoteFileHandler():
    @staticmethod
    def stream_url(url, chunk_size=8192):
        response = urllib.request.urlopen(url)
        head = response.read(4)
        mode = RemoteFileHandler.determine_ftype(head)
        if mode == 'gzip':
            decompressor = zlib.decompressobj(16 + zlib.MAX_WBITS)
            yield decompressor.decompress(head)
            while stream := response.read(chunk_size):
                if not stream:
                    break
                yield decompressor.decompress(stream)
            yield decompressor.flush()

    @staticmethod
    def fetch_url(url: str) -> BytesIO:
        response = urllib.request.urlopen(url)
        return BytesIO(response.read())

    @staticmethod
    def determine_ftype(head:bytes) -> str:
        if head.startswith(GZIP):
            return "gzip"
        elif head.startswith(ZIP):
            return "zip"


    @staticmethod
    def is_gzip(stream: BytesIO) -> bool:
        magic_start = stream.read(2)
        stream.seek(0)
        return magic_start == b'\x1f\x8b'

    @staticmethod
    def unpack_tar_gz_bio(stream: BytesIO) -> BytesIO:
        with gzip.open(stream, "r") as gz:
            return BytesIO(gz.read())





import os
import sys
import json
from dataclasses import dataclass
from io import BytesIO

from aldepyde.env import ENV
from .utils import _parse_memory, _convert_memory_bytes, _convert_memory_bits


def requires_enabled(func):
    def wrapper(cls, *args, **kwargs):
        if hasattr(cls, "enabled") and cls.enabled and cls._initialized:
            return func(cls, *args, **kwargs)
        else:
            return None
    return wrapper


class CacheManager():
    def __init__(self, path=None, initialize=False):
        if not initialize:
            self._initialized = False
            return
        self._initialized = True
        self.enable()
        self._cache_marker = ".aldepyde_cache"
        self.fingerprint = "adpy."
        if path is None:
            self._path = ENV.get_default_path()
        else:
            self._path = path
        if os.path.exists(self.marker_location()):
            try:
                self.load_manager()
            except json.decoder.JSONDecodeError:
                self.load_defaults()
                self.save_settings()
        else:
            self.load_defaults()
            self.save_settings()

    @requires_enabled
    def set_cache_location(self, directory):
        try:
            os.makedirs(directory, exist_ok=True)
            if not os.path.isfile(os.path.join(directory, self._cache_marker)):
                self.save_settings(os.path.join(directory, self._cache_marker))
        except OSError:
            print(f"Error establishing cache directory at : {directory}", file=sys.stderr)

    @requires_enabled
    def set_max_memory(self, memory: str) -> None:
        self.max_memory = _parse_memory(memory)
        self.save_settings()

    @requires_enabled
    def cache_location(self):
        return self._path

    @requires_enabled
    def _dump_settings(self):
        return f"{self.__dict__}"

    @requires_enabled
    def marker_location(self) -> str:
        return os.path.join(self._path, self._cache_marker)

    # Saves settings to a location. If a path is specified, the results are saved to that location
    # but the cache location is NOT changed. Use self.set_cache_location() instead for this
    @requires_enabled
    def save_settings(self, path=None):
        if path is None:
            path = self.marker_location()
        elif os.path.isdir(path):
            path = os.path.join(path, self._cache_marker)
        with open(path, 'w') as fp:
            # for v in vars(self):
            #     print(v)
            fp.write(json.dumps(vars(self), indent=2))

    @requires_enabled
    def load_manager(self, path=None):
        if path is None:
            path = self.marker_location()
        elif os.path.isdir(path):
            path = os.path.join(path, self._cache_marker)
        with open(path, 'r') as fp:
            settings = json.load(fp)
            for setting in settings:
                self.__dict__[setting] = settings[setting]
        self._path = os.path.dirname(path)

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def _enabled_and_initialized(self):
        return self.enabled and self._initialized

    def is_enabled(self):
        return self.enabled

    @requires_enabled
    def load_defaults(self):
        self._cache_marker = ".aldepyde_cache"
        self.version = "1.0.0"
        self.max_memory = _parse_memory('2gib')
        self.enabled = True

    @requires_enabled
    def _inside(self, f) -> str:
        return os.path.join(self._path, f)

    @requires_enabled
    def peek(self, name):
        files = []
        for file in self.list_cache():
            if name.lower() in file.lower():
                files.append(file)
        return files

    # Requires a filename, not a path. The path will be self._path
    @requires_enabled
    def _is_safe_to_delete(self, filename):
        return (os.path.exists(self.marker_location())
                and (filename in os.listdir(self._path) or filename.startswith(self._path))
                and self.exists(filename)
                and (filename.startswith(self.fingerprint)) or os.path.basename(filename).startswith(self.fingerprint))

    @requires_enabled
    def delete_from_cache(self, filename) -> bool | None:
        if not self._enabled_and_initialized():
            return None
        if self._is_safe_to_delete(filename):
            os.remove(self._inside(filename))
            return True
        return False

    @requires_enabled
    def clear_cache(self) -> None:
        if not self._enabled_and_initialized():
            return None
        for file in os.listdir(self._path):
            if self._is_safe_to_delete(file):
                os.remove(self._inside(file))
                print(f"Deleting {file}")

    # @requires_enabled
    # def get(self, file) -> str | None:
    #     if not self._enabled_and_initialized():
    #         return None
    #     return self._inside(file) if self.exists(file) else None

    # TODO Create a with cache_manager.open(...) setup here
    # def open(self, filename):
    #     pass

    @requires_enabled
    def retrieve(self, filename:str) -> str|None:
        if self.exists(filename):
            return self._inside(filename)
        else:
            return None

    @requires_enabled
    def exists(self, filename):
        return os.path.isfile(self._inside(filename))

    @requires_enabled
    def list_cache(self) -> list:
        all_files = []
        for base, _, paths in os.walk(self._path):
            for path in paths:
                all_files.append(os.path.join(base, path))
        return all_files

    @requires_enabled
    def cache_usage(self, output=None):
        total_size = 0
        for filepath in self.list_cache():  # Yes this jumps from O(n^2) to O(n^3), but It's probably fine. Just don't use a petabyte-sized cache
            total_size += os.path.getsize(filepath)
        if output == 'percentage':
            return total_size / self.max_memory
        elif output == 'bytes':
            return _convert_memory_bytes(total_size)
        elif output == 'bits':
            return _convert_memory_bits(total_size)
        return total_size

    @dataclass
    class EvictionResult():
        success: bool
        deleted: list[str]
        error: Exception | None = None

        @property
        def error_message(self) -> str | None:
            return str(self.error) if self.error is not None else None

    @requires_enabled
    def evict(self, nbytes) -> EvictionResult | None:
        if not self._enabled_and_initialized():
            return None
        cache_list = self.list_cache()
        # Clear entries in cache by age
        cache_list.sort(key=os.path.getatime)
        deleted = []
        while nbytes + self.cache_usage() > self.max_memory:
            if not cache_list:
                return self.EvictionResult(False, deleted)
            file = cache_list.pop(0)
            try:
                if self.delete_from_cache(file):
                    deleted.append(file)
            except FileNotFoundError as error:
                return self.EvictionResult(False, deleted, error)
        return self.EvictionResult(True, deleted)

    @requires_enabled
    def clean(self) -> EvictionResult:
        return self.evict(0)

    @requires_enabled
    def save_to_cache(self, stream, filename) -> bool | None:
        filename = os.path.join(self._path, filename)
        if not self._enabled_and_initialized():
            return None
        self.evict(stream.getbuffer().nbytes)
        with open(filename, "wb") as fp:
            fp.write(stream.read())
        stream.seek(0)
        return True



def is_mmcif(text:str) -> bool:
    for line in text.split('\n'):
        if line.startswith('#'):
            continue
        # First non-comment line must be data_{PDB} code
        if text.strip().startswith('data_'):
            return True
        else:
            return False

def _detect_gz(filepath):
    with open(filepath, 'rb') as fp:
        head = fp.read(2)
        if b"\x1f\x8b" in head:
            return True
    return False



# Replacement for deprecated RandomProtein.py
# Everything recoded and refactored to be up to standard

import random
import math
import requests
import json



# TODO This whole thing needs to be cleaned up to better align with more modern python
class RandomProtein:
    # Hardcode remode_data for now
    def __init__(self, His_Is_Charged=True, Cys_Is_Polar=True, Charged_Is_Polar=True, Distribution="Swiss"):
        self.Analyzer = ProteinStats(His_Is_Charged, Cys_Is_Polar, Charged_Is_Polar)

        self._ACCEPTED_METHODS = ["Random", "Custom", "Builder", "Grouper"]
        self._ACCEPTED_PRESETS = ["Swiss", "CHG"]
        self.His_Is_Charged = His_Is_Charged
        self.Cys_Is_Polar = Cys_Is_Polar
        self.Charged_Is_Polar = Charged_Is_Polar

        self.LoadPresetDistribution(Distribution)

        self._AAs = "ARNDCEQGHILKMFPSTWYV"
        self._POSITIVE = "RK"
        self._NEGATIVE = "ED"
        self._POLAR = "STNQ"
        self._NONPOLAR = "AVILMFYWGP"
        self._CHARGED = "RKED"

        if His_Is_Charged:
            self._CHARGED += "H"
            self._POSITIVE += "H"
        else:
            self._POLAR += "H"

        if Cys_Is_Polar:
            self._POLAR += "C"
        else:
            self._NONPOLAR += "C"

        if Charged_Is_Polar:
            self._POLAR += self._POSITIVE
            self._POLAR += self._NEGATIVE

    def GetNP(self):
        return self. _NONPOLAR



    def LoadPresetDistribution(self, preset="Swiss"):
        if preset.upper() == "Swiss".upper():
            _stream = pkg_resources.resource_stream("aldepyde", 'data/Swiss_Prot.data')
            self.distribution = json.load(_stream)['Amino Acid Distribution']
        elif preset.upper() == "CHG":
            _stream = pkg_resources.resource_stream("aldepyde", 'data/CHG.data')
            self.distribution = json.load(_stream)['Amino Acid Distribution']
        else:
            raise InvalidDistribution(
                f"Preset must be one of the following: {self._ACCEPTED_PRESETS}\n\t"
                f"Chosen: {preset}")

    def GetAminoAcids(self):
        return self._AAs

    def SetAminoAcids(self, AAs):
        if isinstance(AAs, str):
            self._AAs = str
        elif isinstance(AAs, list):
            self._AAs = "".join(AAs)

    # TODO add setters for polar, nonpolar, etc.

    def AddAminoAcid(self, AA):
        if AA not in self._AAs:
            self._AAs += AA

    def RemoveAminoAcid(self, AA):
        if AA in self._AAs:
            self._AAs.replace(AA, "")
        if AA in self._POLAR:
            self._POLAR.replace(AA, "")
        if AA in self._CHARGED:
            self._CHARGED.replace(AA, "")
        if AA in self._POSITIVE:
            self._POSITIVE.replace(AA, "")
        if AA in self._NEGATIVE:
            self._NEGATIVE.replace(AA, "")

    def ConfigureAA(self, AA, is_polar=False, is_charged=False, is_positive=False, is_negative=False):
        # if is_charged and not (is_positive or is_negative):
        #     raise ImpossibleSettingsException("The residue must be either positive or negative if it carries a charge")
        if is_negative and is_positive:
            raise ImpossibleSetting("The residue cannot be both positive and negative")

        if is_charged and self.Charged_Is_Polar and AA not in self._POLAR:
            self._POLAR += AA
        if is_charged and AA not in self._CHARGED:
            self._CHARGED += AA
        if is_negative and AA not in self._NEGATIVE:
            self._NEGATIVE += AA
        if is_positive and AA not in self._POSITIVE:
            self._POSITIVE += AA
        if is_polar and AA not in self._POLAR:
            self._POSITIVE += AA
        elif AA not in self._NONPOLAR:
            self._NONPOLAR += AA
        if AA not in self._AAs:
            self._AAs += AA

    def configure(self, His_Is_Charged=True, Cys_Is_Polar=True, Charged_Is_Polar=True):
        self.__init__(His_Is_Charged, Cys_Is_Polar, Charged_Is_Polar)

    def _load(self, json_str):
        self.distribution = self._NormalizeValues(json.loads(json_str)["Amino Acid Distribution"])

    def _NormalizeValues(self, dic):
        total = 0
        for key in dic.keys():
            total += float(dic[key])
        for key in dic.keys():
            dic[key] = float(dic[key]) / total
        return dic

    def LoadDistributionFromFile(self, dist_json_path):
        with open(dist_json_path, "r") as fp:
            self._load(fp.read())

    def LoadDistributionFromURL(self, URL):
        response = requests.get(URL)
        if response.status_code == 200:
            self._load(json.dumps(response.json()))
        else:
            raise ConnectionError(f"Invalid url: {URL}")

    def GenerateProtein(self, length, batch_size=1, method="Random", percent_polar=None,
                        percent_charged=None, final_charge=None, charge_range=None,
                        timeout=30, max_attempts=None, verbose=False, cleanup=False, crash_at_bad_settings=False):

        self.verbose = verbose
        # TODO make it so max_attempts is actually used
        if not self.VerifySettings(length, batch_size=batch_size, method=method, percent_polar=percent_polar,
                                   percent_charged=percent_charged, final_charge=final_charge,
                                   charge_range=charge_range,
                                   max_attempts=max_attempts, crash=crash_at_bad_settings):
            return None

        method = method.upper()
        ret_list = []
        for _ in range(batch_size):
            sequence = "$" * length
            if method == "Random".upper():
                # sequence = self._Random(length)
                attempt = 0
                while not self._ValidateSequence(sequence, percent_polar,
                                                 percent_charged, final_charge, charge_range):
                    sequence = self._Random(length)
                    if verbose:
                        attempt += 1
                        print(f"\rAttempt: {attempt}", end="")
            elif method == "Custom".upper():
                sequence = self._Custom(length)
            elif method == "Builder".upper():
                sequence = self._Builder(length, percent_charged, percent_polar, final_charge)
            elif method == "Grouper".upper():
                attempt = 0
                while not self._ValidateSequence(sequence, percent_polar,
                                                 percent_charged, final_charge, charge_range):
                    sequence = self._Grouper(length, percent_polar, percent_charged, final_charge)
                    if verbose:
                        attempt += 1
                        print(f"\rAttempt: {attempt}", end="")

            if cleanup:
                if verbose:
                    print("\nCleaning result sequence")
                sequence = self._CleanPolar(sequence, percent_polar)
            ret_list.append(sequence)

        if verbose:
            for result in ret_list:
                print(
                    f"Result: {result}\nPercent Polar: {self.Analyzer.PercentPolar(result)}\n"
                    f"Percent Charged: {self.Analyzer.PercentCharged(result)}"
                    f"\nPercent NonPolar: {self.Analyzer.PercentNonPolar(result)}"
                    f"\nTotal Charge: {self.Analyzer.GetCharge(result)}\n")

        return ret_list

    def _Random(self, length):
        sequence = ""
        for i in range(length):
            sequence += self._AAs[random.randrange(0, len(self._AAs))]
        return sequence

    def _Custom(self, length):
        items = list(self.distribution.keys())
        probabilities = list(self.distribution.values())
        return "".join(random.choices(items, weights=probabilities, k=length))

    def _Builder(self, length, percent_charged, percent_polar, final_charge):
        sequence = []
        pc = percent_charged
        pp = percent_polar

        percent_unclear = 0

        unclear = ""

        if percent_charged is None:
            percent_charged = 0
            unclear += self._CHARGED
        if percent_polar is None:
            percent_polar = 0
            unclear += self._POLAR
        unclear += self._NONPOLAR

        u = set()
        for c in unclear:
            u.add(c)
        unclear = ""
        for c in u:
            unclear += c

        if pc is None or pp is None:
            percent_unclear = 1 - (percent_polar + percent_charged)

        # sequence = ["X"] * length
        for i in range(math.ceil(length * percent_polar)):
            sequence.append(random.choice(self._POLAR))
        for i in range(math.ceil(length * percent_polar),
                       math.ceil(length * percent_polar) + math.ceil(
                           length * percent_charged) - self.Analyzer.NumCharged(sequence)):
            sequence.append(random.choice(self._CHARGED))

        if percent_unclear < 0:
            for _ in range(length * percent_unclear):
                sequence.append(random.choice(self._NONPOLAR))

        if final_charge is not None:
            disparity = final_charge - self.Analyzer.GetCharge(sequence)
            ind = 0
            while not (disparity == 0 or disparity == -1 or disparity == 1) and ind < len(sequence):
                if disparity > 0 and sequence[ind] in self._NEGATIVE:
                    sequence[ind] = random.choice(self._POSITIVE)
                    disparity -= 2
                if disparity < 0 and sequence[ind] in self._POSITIVE:
                    sequence[ind] = random.choice(self._NEGATIVE)
                    disparity += 2
                ind += 1

            # if not ind < len(sequence):
            while not (disparity == 0):
                if disparity == 0 or len(sequence) >= length:
                    break
                elif disparity > 0:
                    sequence += random.choice(self._POSITIVE)
                    disparity -= 1
                else:
                    sequence += random.choice(self._NEGATIVE)
                    disparity += 1

        while len(sequence) < length:
            sequence.append(random.choice(self._NONPOLAR))
        random.shuffle(sequence)
        return "".join(sequence)

    def _Grouper(self, length, percent_polar, percent_charged, final_charge):
        if percent_polar is None:
            percent_polar = 0
        if percent_charged is None:
            percent_charged = 0
        if final_charge is None:
            final_charge = 0

        if final_charge != 0 and percent_charged == 0:
            raise ImpossibleSetting("If final_charge is nonzero, percent_charged must also be nonzero")
        p_pos = 0
        if final_charge > 0:
            p_neg = (1 - (final_charge / (percent_charged * length))) / 2
            p_pos = 1 - p_neg
        # else:
        elif final_charge < 0:
            p_pos = (1 - abs(final_charge) / (percent_charged * length)) / 2

        if self.Charged_Is_Polar:
            percent_charged = percent_charged + percent_polar * (len(self._CHARGED) / (len(self._POLAR)))
            unique = "".join([c for c in self._POLAR if c not in self._CHARGED])
            percent_polar = percent_polar - percent_polar * (len(unique) / (len(self._POLAR)))

        sequence = ""
        for i in range(length):
            r = random.random()
            if r < percent_polar:
                sequence += self._POLAR[random.randrange(0, len(self._POLAR))]
            elif percent_polar < r and r < percent_polar + percent_charged:
                r2 = random.random()
                if p_pos > r2:
                    sequence += self._POSITIVE[random.randrange(0, len(self._POSITIVE))]
                else:
                    sequence += self._NEGATIVE[random.randrange(0, len(self._NEGATIVE))]
            else:
                sequence += self._NONPOLAR[random.randrange(0, len(self._NONPOLAR))]
        return sequence

    def _ValidateSequence(self, sequence, percent_polar, percent_charged, total_charge, charge_range):
        if '$' in sequence:
            return False

        if charge_range is None:
            charge_range = 0

        # Polar condition

        if percent_polar is None:
            polar = True
        elif percent_polar < 0:
            polar = True
        else:
            polar = self.Analyzer.PercentPolar(sequence) >= percent_polar
        # Charge condition

        if percent_charged is None:
            charge = True
        elif percent_charged < 0:
            charge = True
        else:
            charge = self.Analyzer.PercentCharged(sequence) >= percent_charged
        # Charge total condition
        charge_t = False
        if total_charge is None:
            charge_t = True
        else:
            charge_t = self.Analyzer.GetCharge(sequence) <= (total_charge + charge_range) \
                       and self.Analyzer.GetCharge(sequence) >= (total_charge - charge_range)

        return (polar and charge and charge_t)

    def VerifySettings(self, length, batch_size, method, percent_polar,
                       percent_charged, final_charge, charge_range, max_attempts, crash=True):

        disregard_charge = False
        if final_charge is None:
            disregard_charge = True
        if batch_size <= 0:
            raise ImpossibleSetting("Batch size must be larger than 0")

        if (max_attempts is not None) and (max_attempts < 0 or not isinstance(max_attempts, int)):
            if crash:
                raise ImpossibleSetting("The number of specified max attempts must be a positive integer or None")
            return False

        if method not in self._ACCEPTED_METHODS:
            if crash:
                raise ImpossibleSetting(
                    f"Selected method must be one of the following: {self._ACCEPTED_METHODS}.\n\t Chosen: {method}")
            return False

        if not ((percent_charged is None or percent_charged >= 0) and (percent_polar is None or percent_polar >= 0)):
            if crash:
                raise ImpossibleSetting("All provided probabilities must be positive or 0")
            return False

        if not self.Charged_Is_Polar:
            a = 0
            b = 0
            if percent_polar is not None:
                a = percent_polar
            if percent_charged is not None:
                b = percent_charged
            if a + b > 1:
                if crash:
                    raise ImpossibleSetting(
                        "The sum of provided probabilities must be less than or equal to 1 if charged residues are not considered polar")
                return False
        if not disregard_charge and length < abs(final_charge):
            if crash:
                raise ImpossibleSetting(
                    f"The length of the output protein must be less than or equal to the magnitude of the charge ({length} < {final_charge})")
            return False
        return True

        if percent_charged == 0 and final_charge != 0:
            if crash:
                raise ImpossibleSetting(f"A nonzero charge cannot be present if the percent_charged is not 0")
            return False

    def _CleanPolar(self, sequence, percent_polar):
        if percent_polar is None:
            return sequence
        disparity = self.Analyzer.NumPolar(sequence) - math.floor(len(sequence) * percent_polar)
        ind = 0
        s = list(sequence)
        while disparity > 0 and ind < len(s):
            if (s[ind] in self._POLAR) and (s[ind] not in self._CHARGED):
                s[ind] = random.choice(self._NONPOLAR)
                disparity -= 1
            ind += 1
        return "".join(s)


import sys
import time

# I'm sorry tqdm. I love you too :(
# This is NOT meant to replace tqdm. Only bother using this if you are on
# a restrictive system that doesn't allow 3rd party resources
class ProgressBar():
    def __init__(self, itb=None, ascii=False, blength=30, total=None, minupdate=.1, fp=sys.stdout):
        self.itb = self.iterize(itb)
        self.ascii = ascii
        self.blength = blength
        self.minupdate = minupdate
        self._ascii_char = self.default_ascii()
        self._uni_char = self.default_uni()
        self._empty = self.default_empty()
        self._progress_printer = self._print_bar if total != 0 else self._print_current
        self.fp = fp
        self.description = ""
        self.current = 0
        self.start = -1


        if total is None:
            try:
                self.total = len(self.itb)
            except TypeError:
                self.total = 0
        else:
            self.total = total

    def load(self, itb):
        self.itb = self.iterize(itb)

    def set_total(self, total):
        self.total = total

    def __iter__(self):
        it = self.itb
        minupdate = self.minupdate
        self.start = last = time.perf_counter()
        current = 0
        pbar_type = self._print_current if self.total == 0 else self._print_bar
        fp = self.fp
        for item in iter(it):
            yield item
            current += 1
            self.current = current
            now = time.perf_counter()
            if now - last >= minupdate:
                last = now
                fp.write("".join(["\r", self.description, pbar_type(current)]))
        fp.write("".join(["\r", self.description , pbar_type(current)]))
        fp.write('\n')

    def update(self, current=1):
        self.current += 1
        if self.start == -1:
            self.start = time.perf_counter()
        if self.total == 0:
            self.sp.write("".join(["\r", self.description ,self._print_current(self.current)]))
        else:
            self.fp.write("".join(["\r",self.description ,self._print_bar(self.current)]))

    def _print_current(self, current:int) -> str:
        return "".join([str(current), self._get_meta_current(current)])

    def _get_meta_current(self, current):
        elapsed = time.perf_counter() - self.start
        eh, er = divmod(elapsed, 3600)
        em, es = divmod(er, 60)
        eh = int(eh)
        em = int(em)
        es = int(es)
        rate = round(current / elapsed)
        elp = "".join([str(eh).zfill(2), ':', str(em).zfill(2), ':', str(es).zfill(2)])
        return "".join(['|', '|ELP ' , elp, '|', str(rate), ' itr/sec'])

    def _print_bar(self, current:int) -> str:
        completed = current / self.total
        cycle = self._ascii_char if self.ascii else self._uni_char
        if len(cycle) == 1:
            return "".join([cycle[0]*round(completed*self.blength), self._empty*(self.blength - round(completed*self.blength)), self._get_meta_pbar(current)])
        else:
            full = min(len(cycle[-1] * int(completed * self.blength)), self.blength)
            empty = self.blength - full - 1
            bridge_index = round(((self.blength * completed)%1) * (len(cycle) - 1))
            bridge = "" if empty < 0 else cycle[bridge_index]
            return "".join( ['[', cycle[-1] * full, bridge, self._empty*empty, self._get_meta_pbar(current), ']'])


    def _get_meta_pbar(self, current):
        elapsed = time.perf_counter() - self.start
        eh, er = divmod(elapsed, 3600)
        em, es = divmod(er, 60)
        eh = int(eh)
        em = int(em)
        es = int(es)
        rate = round(current / elapsed)
        eta = (self.total - current) / (rate + 0.00001)
        h, r = divmod(eta, 3600)
        m, s = divmod(r, 60)
        h = int(h)
        m = int(m)
        s = int(s)
        rs = "".join([str(h).zfill(2), ':', str(m).zfill(2), ':', str(s).zfill(2)])
        elp = "".join([str(eh).zfill(2), ':', str(em).zfill(2), ':', str(es).zfill(2)])
        return "".join(['|', str(round(100*current/self.total)), '%|ETA ', rs, '|ELP ' , elp, '|', str(rate), ' itr/sec'])


    def iterize(self, itb):
        if isinstance(itb, int):
            return range(itb)
        else:
            try:
                iter(itb)
                return itb
            except TypeError:
                sys.stderr.write("The object loaded into the progress bar is not iterable")
                return None

    def default_empty(self):
        return '-'

    def set_empty(self, char: str):
        if len(char) != 1:
            raise ValueError("The empty character must only be one character")
        self._empty = char

    def default_ascii(self):
        return ('@')

    def set_ascii(self, progression : tuple):
        self._ascii_char = progression

    def default_uni(self):
        return ('░', '▒', '▓', '█')

    def set_uni(self, progression : tuple):
        self._uni_char = progression

    def set_description(self, desc):
        self.description = desc




class ProgressBar_dep():
    def __init__(self, itb=None, force_ascii=False, force_spinner=False, bar_length=10):
        self.itb = None
        self.iterator = None
        self.current = 0
        self.total = 0
        self.bar_length = bar_length
        self.description = ""
        self._force_ascii = force_ascii
        self.empty = ' '
        self.spinner = ['\\', '|', '/', '-']
        # TODO Make _quarters adjustable
        self._quarters = ['░', '▒', '▓', '█']
        self.set_quarters(self._quarters)
        self._force_spinner=force_spinner
        self.set_force_ascii(self._force_ascii)
        self.times = []
        if itb is None:
            return
        else:
            self.load(itb)

    # Mostly for debugging
    def set_force_spinner(self, force_spinner):
        self._force_spinner = force_spinner

    # Mostly for debugging
    def set_force_ascii(self, force_ascii):
        self._force_ascii = force_ascii
        if not force_ascii:
            self.spinner = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']
        else:
            self.spinner = ['\\', '|', '/', '-']

    def set_quarters(self, quarters):
        self._quarters = [self.empty] + quarters

    def set_total(self, total):
        self.total = total

    def load(self, itb):
        if itb is not None:
            if isinstance(itb, int):
                itb = range(itb)
            try:
                iter(itb)
                self.current = 0
                self.description = ""
                self.itb = itb
                self.iterator = iter(self.itb)
                try:
                    self.total = len(itb)
                except TypeError:
                    self.total = 0

            except TypeError:
                print("The object loaded into the progress bar is not iterable", file=sys.stderr)


    def update(self, updates : int):
        self.current += updates

    def set_description(self, description):
        if description is None:
            description = ""
        self.description = description

    # TODO Toggle ETA
    def eta(self):
        if len(self.times) == 2:
            try:
                seconds = (self.times[1] - self.times[0]) * (self.total - self.current)
                hours, remainder = divmod(seconds, 3600)
                minutes, sec = divmod(remainder, 60)
                tstr = ""
                if hours >= 1:
                    tstr += f"{int(hours)}h "
                if minutes >= 1:
                    tstr += f"{int(minutes)}m "
                tstr += f"{round(sec)}s "
                return f" {{{tstr}}}"
            except ZeroDivisionError:
                return ""
        else:
            return ""

    def push_time(self, t):
        while len(self.times) >= 2:
            self.times.pop(0)
        self.times.append(t)

    def _progress_as_bar(self) -> str:
        return ""
        try:
            current_progress = self.current / self.total
        except ZeroDivisionError:
            return ""
        if self._force_ascii:
            current_progress = round(current_progress * self.bar_length)
            return '*' * current_progress + '-' * (self.bar_length - current_progress)
        else:
            current_progress = current_progress * self.bar_length
            whole_length = int(current_progress // 1)
            whole_portion = self._quarters[-1] * whole_length
            bp = round((current_progress%1) * 4)
            bridge = self._quarters[bp]
            decimal_portion = '-' * (self.bar_length - whole_length - 1)
            return whole_portion + bridge + decimal_portion

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.push_time(time())
            item = next(self.iterator)

            self.current += 1
            d = " | " + self.description if len(self.description) != 0 else ""
            return item

            progress_statement = f"{self.current}/{self.total}" if self.total != 0 and not self._force_spinner else f"{self.current} {self.spinner[self.current % len(self.spinner)]}"
            sys.stdout.write('\r' + self._progress_as_bar() + ' ' + progress_statement + d + self.eta())
            sys.stdout.flush()
            return item

        except StopIteration:
            print()
            raise


from typing import List, Tuple, Optional, Dict
from dataclasses import dataclass, field
from ..biomolecule.structure import Atom, Residue, Chain, Model, Structure


def catagorize_lines(text, *args:str) -> dict:
    catagories = {k:list() for k in args}
    for line in text.split('\n'):
        if line.startswith(args):
            match = next(p for p in args if line.startswith(p))
            catagories[match].append(line)
    return catagories

# PDB documentation guarantees only 1 HEADER per file
def get_title_from_header(header):
    return header[0][62:66] if len(header) != 0 else None

def _slice(s, start, end):
    if start >= len(s):
        return ""
    return s[start:end].strip()

def _parse_atom(s:str) -> Optional[Atom]:
    serial = _slice(s, 6, 11)
    atom_name = _slice(s, 12, 16)
    altLoc = _slice(s, 16, 17)
    x = _slice(s, 30, 38)
    y = _slice(s, 38, 46)
    z = _slice(s, 46, 54)
    try:
        x = float(x)
        y = float(y)
        z = float(z)
    except ValueError:
        return None
    occupancy = _slice(s, 54, 60)
    try:
        occupancy = float(occupancy)
    except ValueError:
        return None
    tempFactor = _slice(s, 60, 66)
    try:
        tempFactor = float(tempFactor)
    except ValueError:
        return None
    element = _slice(s, 76, 78)
    charge = _slice(s, 78, 80)
    return Atom(serial=serial,
                name=atom_name,
                altloc=altLoc,
                coord=(x, y, z),
                occupancy=occupancy,
                b_factor=tempFactor,
                element=element,
                charge=charge,
                is_het=s.startswith('HETATM'))

def _parse_pdb(text:str) -> Optional[Structure]:
    special_lines = catagorize_lines(text, 'HEADER', "TITLE", "REMARK 350")
    structure_name = get_title_from_header(special_lines['HEADER'])
    title = "".join(special_lines['TITLE']) # TODO clean this up
    structure = Structure(id=structure_name)
    # structure = Structure(id=structure_name, title=title)
    lines = text.split('\n')
    current_model = None
    chains:Dict[str, Chain] = {}
    residues:Dict[str, Residue] = {}
    # structure.add_model(current_model)
    for line in lines:
        if line.startswith('MODEL'):
            model_index = _slice(line, 10, 14)
            current_model = Model(model_index)
            structure.add_model(current_model)
            chains = {}
            residues = {}
        if line.startswith('ENDMDL'):
            chains = {}
            residues = {}
        if line.startswith(('ATOM', 'HETATM')):
            if current_model is None:
                current_model = Model('1')
                chains: Dict[str, Chain] = {}
                residues: Dict[str, Residue] = {}
                structure.add_model(current_model)
            atom = _parse_atom(line)
            if atom is None:
                continue
            chain_id = _slice(line, 21, 22)
            if chain_id not in chains:
                chains[chain_id] = Chain(id=chain_id)
                current_model.add_chain(chains[chain_id])

            resName = _slice(line, 17, 20)
            resSeq = _slice(line, 22, 26)
            # resSeq = int(resSeq)

            i_code = _slice(line, 26, 27)
            if resSeq not in residues:
                residues[resSeq] = Residue(resName, resSeq, i_code)
                chains[chain_id].add_residue(residues[resSeq])
            residues[resSeq].add_atom(atom)
    return structure


from importlib.resources import files
from typing import Literal
import random
import json
from aldepyde.data.data_exceptions import AmbiguousTokenException, DataAccessException, BadCategoryException

_PACKAGE = 'aldepyde.data'

#TODO Clean up your exceptions. This is a tad excessive as is
class DataStore():
    def __init__(self):
        path = files(_PACKAGE) / f"chemistry.json"
        self._raw = json.loads(path.read_text(encoding='utf-8'))

    def get_raw(self):
        return self._raw

    def get(self, *args:str, default=None):
        current_level = self.get_raw()
        try:
            for arg in args:
                current_level = current_level[arg]
            return current_level
        except KeyError as e:
            return default

    def require(self, *args:str):
        current_level = self._raw
        attempted = []
        try:
            for arg in args:
                attempted.append(arg)
                current_level = current_level[arg]
            return current_level
        except KeyError as e:
            raise DataAccessException(f"\n\nInvalid data access path: {' | '.join(attempted)}\nValid "
                                      f"keys at this level are:\n {', '.join(list(current_level.keys()))}")

    def has(self, *args:str) -> bool:
        current_level = self._raw
        try:
            for arg in args:
                current_level = current_level[arg]
            return True
        except KeyError as e:
            return False

    def is_value(self, *args:str) -> bool:
        current_level = self._raw
        attempted = []
        try:
            for arg in args:
                attempted.append(arg)
                current_level = current_level[arg]
            return not isinstance(current_level, dict)
        except KeyError as e:
            raise DataAccessException(f"\n\nInvalid data access path: {' | '.join(attempted)}\nValid "
                                      f"keys at this level are:\n {', '.join(list(current_level.keys()))}")

    def is_key(self, *args:str) -> bool:
        return not self.is_value(*args)

    def keys(self, *args:str, strict=True) -> list | None:
        current_level = self._raw
        attempted = []
        try:
            for arg in args:
                attempted.append(arg)
                current_level = current_level[arg]
            if isinstance(current_level, dict):
                return list(current_level.keys())
            else:
                return None
        except KeyError as e:
            raise DataAccessException(f"\n\nInvalid data access path: {' | '.join(attempted)}\nValid "
                                      f"keys at this level are:\n {', '.join(list(current_level.keys()))}")

    def all_residues(self, res_type) -> list[str]:
        category = self.categorize(res_type)
        return list(self.require(category).keys())


    canon_aa = 'amino_acid'
    canon_dna = 'dna'
    canon_rna = 'rna'
    canon_ele = 'element'
    categories = {
        'aa' : canon_aa,
        'a' : canon_aa,
        'amino acid' : canon_aa,
        'amino_acid' : canon_aa,
        'dna' : canon_dna,
        'd' : canon_dna,
        'rna' : canon_rna,
        'r' : canon_rna,
        'element' : canon_ele,
        'atom' : canon_ele,
        'e' : canon_ele
    }

    def _category_exception_grouper(self):
        d = {}
        for k, v in self.categories.items():
            d[v] = d.get(v, []) + [k]
        ret_str = ""
        for k, v in d.items():
            ret_str += f"\t{v} -> {k}\n"
        return ret_str


    def categorize(self, res_type:str) -> str:
        try:
            normalized = self.categories[res_type.lower().strip()]
            return normalized
        except KeyError:
            raise BadCategoryException(f"\n\nUnresolvable residue type {res_type}\nAvailable residue types:\n{self._category_exception_grouper()}")


    def normalize_residue(self, name:str, res_type:str='aa') -> str | None:
        if res_type is None:
            return None
        normalized = self.require('map', self.categorize(res_type), name.lower().strip())
        if isinstance(normalized, list):
            raise AmbiguousTokenException(f'\n\nUnable to resolve normalize name: {name} to a single residue\n{name} -> {normalized}\n'
                                          f'Use resolve_residues() when ambiguous tokens are expected.')
        return normalized


    def resolve_residues(self, name:str, res_type:str='aa')->tuple:
        normalized = self.require('map', self.categorize(res_type), name.lower().strip())
        if isinstance(normalized, list):
            return tuple(normalized)
        return (normalized,)

    def get_entry(self, name, res_type='aa', strict=True, resolve_ambiguous:Literal['none', 'first', 'last', 'random']="none") -> dict|list[dict]|None:
        category = self.categorize(res_type)
        if strict:
            name = self.normalize_residue(name, category)
            return self.require(category, name)
        else:
            name = self.resolve_residues(name, category)
            if len(name) == 1:
                return self.get(category, name[0])
            elif len(name) == 0:
                return None
            if resolve_ambiguous == 'none':
                return self.get(category, name[0])
            elif resolve_ambiguous == 'first':
                return self.get(category, name[0])
            elif resolve_ambiguous == 'last':
                return self.get(category, name[-1])
            elif resolve_ambiguous == 'random':
                return self.get(category, name[random.randint(0, len(name) - 1)])
            else:
                raise ValueError(f"Invalid value {resolve_ambiguous} for parameter resolve_ambiguous=['none', 'first', 'last', 'random']")

    def get_entries(self, name, res_type='aa') -> tuple[dict]:
        category = self.categorize(res_type)
        name = self.resolve_residues(name, category)
        ret_lst = []
        for n in name:
            ret_lst.append(self.get(category, n))
        return tuple(ret_lst)


    def _map(self, category):
        return self.require('map', self.categories[category])

    def is_aa(self, name:str) -> bool:
        return name.lower() in self.require('map', 'amino_acid')

    def is_dna(self, name:str) -> bool:
        return name.lower() in self.require('map', 'dna')

    def is_rna(self, name:str) -> bool:
        return name.lower() in self.require('map', 'rna')

    def is_element(self, name:str) -> bool:
        return name.lower() in self.require('map', 'element')

    # What res_types does a token exist in?
    def possible_categories(self, token):
        return [key for key in self.require('map') if token.lower() in self.require('map')[key].keys()]

    def get_all_possible_entries(self, name, strict=True) -> tuple:
        entries = list()
        for res_type in self.possible_categories(name):
            current_name = self.normalize_residue(name, res_type=res_type)
            if strict:
                entries.append(self.require(res_type, current_name))
            else:
                entries.append(self.get(res_type, current_name))
        return tuple(entries)

    def one_letter(self, res, res_type='aa', strict=True):
        return self.get_entry(name=res, res_type=res_type, strict=strict)['1code']

    def dna_pdb_code(self, res, strict=True):
        return self.get_entry(res, self.canon_dna, strict=strict)['pdb_code']

    def three_letter(self, res, strict=True):
        return self.get_entry(res,self.canon_aa, strict=strict)['3code']

    def atom_list(self, res, res_type='aa', strict=True):
        return self.get_entry(name=res, res_type=res_type, strict=strict)['atoms']

    def centroid_atoms(self, res, res_type='aa', strict=True) -> list:
        return self.get_entry(name=res, res_type=res_type, strict=strict)['centroid']

    def formula(self, res, res_type='aa', strict=True):
        return self.get_entry(name=res, res_type=res_type, strict=strict)['formula']

    def hydropathy(self, res, scale:Literal['kd'], strict=True) -> float:
        return self.get_entry(name=res, res_type='aa', strict=strict)['hydropathy'][scale]

    def helical_prop(self, res, strict=True) -> float:
        return self.get_entry(name=res, res_type='aa', strict=strict)['std-alpha-prop']

    # TODO Handle unnatural elements
    def atomic_mass(self, name):
        return self.get_entry(name=name, res_type=self.canon_ele, strict=True)['molar mass']

    def atomic_number(self, name):
        return self.get_entry(name=name, res_type=self.canon_ele,strict=True)['number']

    def citations(self):
        return self.require("Sources")

    def thanks(self):
        return self.require("Special Thanks and Acknowledgment")


_STORE = None

def get_store():
    global _STORE
    if _STORE is None:
        _STORE = DataStore()
    return _STORE

def del_store():
    global _STORE
    _STORE = None


from abc import ABC, abstractmethod
from typing import Iterable, List

from .polymer_specs import PolymerClassifier
from .polymer_constraints import PolymerConstraints
import random
import os
from aldepyde.distribution.distribution import Distribution, load
from aldepyde.distribution.distributions.standards import distribution_head
import json
import math
from itertools import combinations


class ResidueGenerator(ABC):
    def __init__(self, classifier:PolymerClassifier):
        self.classifier = classifier

    def _generate(self):
        pass

    # Just pure, random generation. Nothing fancy
    def random(self, length:int, n:int=1) -> list:
        alphabet = list(self.classifier.alphabet)
        sequences = []
        for i in range(n):
            sequences.append("".join(random.choices(alphabet, k=length)))
        return sequences

    def _list_remove(self, l, element) -> list:
        new_list = []
        for val in l:
            if val != element:
                new_list.append(val)
        return new_list


    # Actual magic
    def _gosper(self, n, k):
        if k < 0 or k > n:
            return
        x = (1 << k) - 1
        limit = 1 << n
        while x < limit:
            yield format(x, f"0{n}b")
            c = x & -x
            r = x + c
            x = (((r ^ x) >> 2) // c) | r


    def possible_mutations(self, sequence:str|Iterable[str], n=1, mutable:Iterable|None=None, alphabet:Iterable|None=None, strict:bool=False):
        r_alphabet = list(self.classifier.alphabet) if alphabet is None else alphabet
        mutable_sequence = sequence if mutable is None else "".join([x for x in sequence if x in mutable])
        total = 0
        for binary in self._gosper(len(mutable_sequence), n):
            if not binary.count('1') == n:
                continue
            subproduct = 1
            for index in [i for i, digit in enumerate(binary) if digit == '1']:
                subproduct *= len(r_alphabet) if mutable_sequence[index] not in r_alphabet else len(r_alphabet) - 1
            total += subproduct
        return total


    def random_mutation(self, sequence:str, k:int=1, n:int=1, mutable:Iterable|None=None, alphabet:Iterable|None=None, strict:bool=False) -> List[str]:
        r_alphabet = list(self.classifier.alphabet) if alphabet is None else alphabet
        # Assume an iterable
        rets = set()
        allowed_indeces = range(0, len(sequence)) if mutable is None else [x for x, val in enumerate(sequence) if
                                                                           val in mutable]
        max_mutations = self.possible_mutations(sequence, n=n, mutable=mutable, alphabet=alphabet)
        target_number = min([k, max_mutations])
        while len(rets) < target_number:
            s = list(sequence)
            if strict:
                k_n = n
            else:
                k_n = n if n <= len(allowed_indeces) else len(allowed_indeces)
            for index in  random.sample(allowed_indeces, k_n):
                new_res = random.choice(self._list_remove(r_alphabet, s[index]))
                s[index] = new_res
            rets.add("".join(s))
        return list(rets)

    def random_from_distribution(self, distribution: os.PathLike|str|dict[str, float]|Distribution, length:int|tuple[int,int], n:int=1, exclude:tuple=()) -> list:
        if isinstance(distribution, str) and os.path.isfile(distribution): # Given a file
            with open(distribution) as fp:
                data = json.load(fp)
                if distribution_head not in data.keys():
                    raise ValueError(f"Key '{distribution_head}' must be in the top level of your provided file")
            usable_distribution = Distribution(None, data[distribution_head])
        elif isinstance(distribution, str) and not os.path.isfile(distribution): # Given a preload
            usable_distribution = load(distribution)
        elif isinstance(distribution, Distribution): # Given a distribution
            usable_distribution = distribution
        elif isinstance(distribution, dict): # Given a dictionary
            usable_distribution = Distribution(None, distribution)
        else:
            raise ValueError('Invalid distribution')
        for r in exclude:
            usable_distribution.eliminate_entry(r)
        if not isinstance(length, int):
            size = lambda: random.randint(length[0], length[1])
        else:
            size = lambda: length
        usable_distribution.normalize_map()
        sequences = []
        for _ in range(n):
            sequence = random.choices(population=list(usable_distribution.frequency_map.keys()), weights=list(usable_distribution.frequency_map.values()), k=size())
            sequences.append("".join(sequence))
        return sequences

    def random_mutation_from_distribution(self, sequences:str|Iterable[str], distribution: os.PathLike|str|dict[str, float]|Distribution, k:int=1, exclude:tuple=()) -> list:
        if isinstance(sequences, str):
            all_values = [sequences]
        else:
            all_values = sequences

        if isinstance(distribution, str) and os.path.isfile(distribution): # Given a file
            with open(distribution) as fp:
                data = json.load(fp)
                if distribution_head not in data.keys():
                    raise ValueError(f"Key '{distribution_head}' must be in the top level of your provided file")
            usable_distribution = Distribution(None, data[distribution_head])
        elif isinstance(distribution, str) and not os.path.isfile(distribution): # Given a preload
            usable_distribution = load(distribution)
        elif isinstance(distribution, Distribution): # Given a distribution
            usable_distribution = distribution
        elif isinstance(distribution, dict): # Given a dictionary
            usable_distribution = Distribution(None, distribution)
        else:
            raise ValueError('Invalid distribution')
        for r in exclude:
            usable_distribution.eliminate_entry(r)
        usable_distribution.normalize_map()

        rets = []
        for sequence in all_values:
            s = list(sequence)
            for index in  random.sample(range(0, len(sequence)), k):
                while True:
                    selection = random.choices(population=list(usable_distribution.frequency_map.keys()),
                                   weights=list(usable_distribution.frequency_map.values()), k=1)[0]
                    if selection != s[index]:
                        s[index] = selection
                        break
            rets.append("".join(s))
        return rets


    @abstractmethod
    def generate(self, length:int, k:int, constraints:PolymerConstraints) -> List[str]:
        expected_per_class = constraints.specs.get_expected(length)
        class_distinctions = constraints.specs.categories
        results = []
        for _ in range(k):
            new_sequence = []
            for key in expected_per_class:
                new_sequence += random.choices(tuple(class_distinctions[key]), k=expected_per_class[key])
            random.shuffle(new_sequence)
            results.append("".join(new_sequence))
        return results
        # for _ in range(k):
        #     for key in expected:




from dataclasses import dataclass
from importlib.resources import files
from typing import Mapping
import json
import copy

@dataclass
class Distribution():
    name: str|None
    frequency_map: dict[str, float]
    source: str | None = None
    version: str | None = None

    def __post_init__(self): # Create a new pointer to the frequency
        self.frequency_map = dict(self.frequency_map)

    def normalize_map(self):
        total = sum(self.frequency_map.values())
        for key in self.frequency_map:
            self.frequency_map[key] /= total
        return self

    def eliminate_entry(self, key):
        if key in self.frequency_map.keys():
            self.frequency_map.pop(key)



_PACKAGE = "aldepyde.distribution.distributions"
def load(name: str) -> Distribution:
    path = files(_PACKAGE) / f"{name}.json"
    if not path.is_file():
        available = sorted(p.stem for p in files(_PACKAGE).iterdir() if p.suffix == '.json')
        raise ValueError(f'Unknown distribution "{name}". Available: {available}')

    raw = json.loads(path.read_text(encoding='utf-8'))
    return Distribution(
        name=name,
        frequency_map=raw.get('distribution'),
        source=raw.get('source'),
        version=raw.get('version')
    )

def available_distributions() -> list:
    available = sorted(p.stem for p in files(_PACKAGE).iterdir() if p.suffix == '.json')
    return available


# Configurable parent class
from abc import ABC, abstractmethod

class Configurable(ABC):
    @abstractmethod
    def UpdateSettings(self):
        """Update all settings upon loading or updating a _config"""


from aldepyde.biomolecule_old._Atom import Atom as atom

__all__ = ['CreateAtom', 'CreateEnforced', 'CreateDummy']

def CreateAtom(  # General atom
        record_name: str| None = None,
        serial: str | int | None = None,
        name: str | None = None,
        altLoc: str | None = None,
        resName: str | None = None,
        chainID: str | None = None,
        resSeq: str | None = None,
        iCode: str | None = None,
        x: str | float | None = None,
        y: str | float |  None = None,
        z: str | float |  None = None,
        occupancy: str | float |  None = None,
        tempFactor: str | float |  None = None,
        element: str | None = None,
        charge: str | None = None
) -> atom:
    return atom(record_name=record_name, serial=serial, name=name, altLoc=altLoc, resName=resName,
                chainID=chainID, resSeq=resSeq, iCode=iCode, x=x, y=y, z=z,
                occupancy=occupancy, tempFactor=tempFactor, element=element,
                charge=charge)

def CreateEnforced( # The type of atom from the parsers. Enforce what's required by PDB standard
        serial: str,
        name: str,
        altLoc: str,
        resName: str,
        chainID: str,
        resSeq: str,
        iCode: str,
        x: str,
        y: str,
        z: str,
        occupancy: str,
        tempFactor: str,
        record_name: str = None,
        element: str | None = None,
        charge: str | None = None
) -> atom:
    return atom(serial=serial, name=name, altLoc=altLoc, resName=resName,
                chainID=chainID, resSeq=resSeq, iCode=iCode, x=x, y=y, z=z,
                occupancy=occupancy, tempFactor=tempFactor, element=element,
                charge=charge, record_name=record_name)


def CreateDummy( # As the name implies
        record_name: str| None = "HETATM",
        serial: str | int = "-1",
        name: str = "PS",
        altLoc: str = "",
        resName: str = "PSD",
        chainID: str = "$P",
        resSeq: str = "PSD",
        iCode: str = "",
        x: str | float = "0.0",
        y: str | float = "0.0",
        z: str | float = "0.0",
        occupancy: str | float = "0.0",
        tempFactor: str | float = "0.0",
        element: str | None = "PS",
        charge: str | None = None
) -> atom:
    return atom(record_name=record_name, serial=serial, name=name, altLoc=altLoc, resName=resName,
                chainID=chainID, resSeq=resSeq, iCode=iCode, x=x, y=y, z=z,
                occupancy=occupancy, tempFactor=tempFactor, element=element,
                charge=charge)




# # Data cache handler
# import shutil
# import os
# import requests
# import re # We're getting spicy with this
# import gzip
# from io import BytesIO, TextIOWrapper
# import data
# from dataclasses import dataclass
# from aldepyde.env import ENV
#
#
# class _cache_handler():
#     class CachePointerException(Exception):
#         pass
#
#     @dataclass
#     class EvictionResult():
#         success: bool
#         deleted: list[str]
#         error: Exception | None = None
#
#         @property
#         def error_message(self) -> str | None:
#             return str(self.error) if self.error is not None else None
#
#     def __init__(self, enabled: bool=True, path: str=None, max_memory: str="2gib", null: bool=False):
#         self.null = null
#         if null:
#             self._enabled = False
#             return
#         self.cache_marker = ".aldepyde_cache"
#         self.version = "1.0.0"
#         if path is None:
#             self._path = ENV.get_default_path()
#         else:
#             self._path = path
#         if os.path.exists(os.path.join(self._path, self.cache_marker)):
#             marker_path = os.path.join(self._path, self.cache_marker)
#             try:
#                 self.load(marker_path)
#             except Exception as e:
#                 self._rectify_marker(marker_path)
#         else:
#             self._enabled = enabled
#             self._max_memory = self._parse_memory(max_memory) # Memory in MiB
#         self.designate_cache(self._path)
#
#     def _rectify_marker(self, marker_path): # Since settings have not been loaded, this MUST come from an environment variable
#         env = os.getenv(ENV.CACHE_REPAIR)
#         print(os.environ)
#         if env is None:
#             raise RuntimeError("No repair policy set. Please set environment variable 'ALDEPYDE_REPAIR_POLICY'"
#                                " to one of the following: ['fail', 'replace', 'backup']")
#         elif env.lower() == "replace": # Just overwrite the old file with defaults
#             self._set_defaults()
#         elif env.lower() == "backup": # Rename the old file to filename.bak
#             marker_path = os.path.join(self._path, self.cache_marker)
#             os.rename(marker_path, marker_path + ".bak")
#             self._set_defaults()
#         elif env.lower() == "fail": # Raise an exception
#             raise RuntimeError(f"The aldepyde module cache has been corrupted.\n"
#                                f"Rerun your program after setting {ENV.CACHE_REPAIR} to either 'replace' or 'backup'"
#                                f"to repair the cache.")
#         else: # Incorrect or unknown command
#             raise ValueError(f"Unkown value set to environment variable {ENV.CACHE_REPAIR}. Valid inputs are ['replace', 'backup', 'fail']\n"
#                              f"\t{ENV.CACHE_REPAIR}={env}")
#
#     def _set_defaults(self):
#         self._enabled = True
#         self._max_memory = self._parse_memory("2gib")
#
#     def load(self, path):
#         with open(path, "r") as fp:
#             settings = data.load(fp)
#         self._enabled = settings['enabled']
#         self._max_memory = settings['max_memory']
#
#     def _save_cache(self):
#         marker = os.path.join(self._path, self.cache_marker)
#         # with open(marker, "r") as fp:
#         #     settings = data.load(fp)
#         settings = {}
#         settings["version"] = self.version
#         settings["enabled"] = self._enabled
#         settings["path"] = self._path
#         settings["max_memory"] = self._max_memory
#         with open(marker, "w") as fp:
#             fp.write(data.dumps(settings, indent=2))
#
#     def designate_cache(self, path) -> None:
#         os.makedirs(self._path, exist_ok=True)
#         self._save_cache()
#         print(os.path.join(path, self.cache_marker))
#         with open(os.path.join(path, self.cache_marker), "w") as fp:
#             fp.write("{}")
#
#     def _verify_cache_directory(self, path: str) -> bool:
#         return os.path.exists(os.path.join(path))
#
#     def _is_safe_to_delete(self, path: str) -> bool:
#         marker_path = os.path.join(path, self.cache_marker)
#         return (
#             os.path.exists(marker_path) and
#             os.path.isdir(path) and
#             os.path.abspath(path) != "/" and
#             os.path.basename(path).startswith("aldepyde_cache")
#         )
#
#     def _get_default_cache_path(self) -> str:
#         return os.path.join(
#             os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), "aldepyde")
#
#
#
#     def set_enabled(self, enabled:bool) -> None:
#         self._enabled = enabled
#         self._save_cache()
#
#     def set_path(self, path:str, cache_policy:str=None) -> None:
#         if cache_policy.lower() == "move":
#             shutil.move(self._path, path)
#         elif cache_policy.lower() == "copy":
#             shutil.copy(self._path, path)
#         self._path = path
#         self._save_cache()
#
#
#     def set_max_memory(self, memory: str) -> None:
#         self._max_memory = self._parse_memory(memory)
#         self._save_cache()
#
#     def _parse_memory(self, memory: str) -> int:
#         ALLOWED_PREFIX = "bkmgt" # No, you don't get to use petabytes
#         full_re = f"[0-9]+[{ALLOWED_PREFIX}]?i?b?"
#         numeric_re = "[0-9]+"
#
#         if memory.isnumeric():
#             memory += "mib"
#         if re.fullmatch(full_re, memory, flags=re.IGNORECASE) is None:
#             raise ValueError(f"Requested memory must be of the following form: {full_re}")
#
#         match = re.match(numeric_re, memory)
#         numeric = int(memory[:match.span()[1]])
#         unit = memory[match.span()[1]:]
#         base = 1024 if "i" in unit else 1000
#         multiple = base**(ALLOWED_PREFIX.index(unit[0].lower()))
#         return numeric * multiple
#
#     def grab_url(self, url: str, filename: str) -> BytesIO:
#         # Return a requested file as a BytesIO stream from a URL or the cache
#         if not self.in_cache(filename):
#             response = requests.get(url)
#             response.raise_for_status()
#             stream_io = BytesIO(response.content)
#             self.save_to_cache(stream_io, filename)
#             with gzip.open(stream_io, "r") as gz:
#                 return BytesIO(gz.read())
#         else:
#             return self.extract_from_cache(filename)
#
#     def clear_cache(self) -> None:
#         if self._is_safe_to_delete(self._path):
#             for p in os.listdir(self._path):
#                 path = os.path.join(self._path, p)
#                 if os.path.isdir(path):
#                     shutil.rmtree(path)
#                 else:
#                     os.remove(path)
#
#     # def SaveCache(self, destination, compress=False):
#     #     if not compress:
#
#
#     def delete_cache(self) -> None:
#         if self._is_safe_to_delete(self._path):
#             shutil.rmtree(self._path)
#
#     def cache_replace(self, nbytes) -> EvictionResult:
#         cache_list = self.list_cache()
#         cache_list.sort(key=os.path.getatime)
#         deleted = []
#         while nbytes + self.cache_usage() > self._max_memory:
#             if not cache_list:
#                 return self.EvictionResult(False, deleted)
#             file = cache_list.pop(0)
#             try:
#                 os.remove(file)
#                 deleted.append(file)
#             except FileNotFoundError as error:
#                 return self.EvictionResult(False, deleted, error)
#         return self.EvictionResult(True, deleted)
#
#
#     def _make_cache(self) -> None:
#         if not os.path.isdir(self._path):
#             os.mkdir(self._path)
#
#
#
#     # TODO This could maybe be expanded to look for compressed/decompressed versions of a file, but that may cause issues later
#     def in_cache(self, filename: str) -> bool:
#         if not self._enabled: # If the cache is disabled, we behave as if the file doesn't exist
#             return False
#         if os.path.exists(os.path.join(self._path, filename)):
#             return True
#         return False
#
#     def extract_from_cache(self, filename:str) -> BytesIO | None:
#         if self.in_cache(filename):
#             _, file_extension = os.path.splitext(filename)
#             if file_extension == ".gz":
#                 with gzip.open(os.path.join(self._path, filename), "rb") as gz:
#                     return BytesIO(gz.read())
#             else:
#                 with open(filename, "rb") as fp:
#                     stream = BytesIO()
#                     stream.write(fp.read())
#                     return stream
#         return None
#
#     def list_cache(self) -> list:
#         all_files = []
#         for base, _, paths in os.walk(self._path):
#             for path in paths:
#                 all_files.append(os.path.join(base, path))
#         return all_files
#
#     def cache_usage(self, percentage: bool=False) -> float | int:
#         total_size = 0
#         for filepath in self.list_cache(): # Yes this jumps from O(n^2) to O(n^3), but It's probably fine. Just don't use a petabyte-sized cache
#             total_size += os.path.getsize(filepath)
#         if percentage:
#             return total_size / self._max_memory
#         return total_size
#
#
#     def save_to_cache(self, stream, filename) -> bool:
#         filename = os.path.join(self._path, filename)
#         self._make_cache()
#         print(stream.getbuffer().nbytes)
#         if not self._enabled:
#             return False
#
#         # Clear entries in cache by age
#         cache_list = self.list_cache()
#         cache_list.sort(key=os.path.getctime)
#         while stream.getbuffer().nbytes + self.cache_usage() > self._max_memory:
#             # print(f"Removing {cache_list[0]}")
#             os.remove(cache_list[0])
#             cache_list.pop(0)
#         if stream.getbuffer().nbytes + self.cache_usage() < self._max_memory:
#             with open(filename, "wb") as fp:
#                 fp.write(stream.read())
#         stream.seek(0)
#         return True
#



# from aldepyde.generators.deprecated.RandomProtein import *
from .protein_generator import ProteinGenerator



# from aldepyde.rand_utils.ProteinClassifier import ProteinClassifier

# __all__ = ['RandomProtein', 'protein_specs.py']

import sys
sys.stderr.write("Note that the `generators` submodule is not yet fully tested and may be unstable\n")


from typing import Optional
from ..biomolecule.structure import Structure
from .parse_pdb import _parse_pdb
from .parse_mmcif import _parse_mmcif
from .sniff import is_mmcif

# This could grow, so alone it goes!
def parse(text:str) -> Optional[Structure]:
    if is_mmcif(text):
        return _parse_mmcif(text)
    else:
        return _parse_pdb(text)



import os
import sys

class ENV():
    CACHE_PATH = "ALDEPYDE_CACHE_DIRECTORY"
    CACHE_REPAIR = "ALDEPYDE_REPAIR_POLICY"
    VERBOSE = "ALDEPYDE_VERBOSE_POLICY"
    APP = "aldepyde"

    @staticmethod
    def set_default_env_vars():
        ENV.set_env(ENV.CACHE_PATH, ENV.get_default_path())
        ENV.set_env(ENV.CACHE_REPAIR, "fail")
        ENV.set_env(ENV.VERBOSE, "false")

    @staticmethod
    def set_env(var, val, force=True):
        if not hasattr(ENV, var):
            raise ValueError(f"{var} is not a valid aldepyde.ENV key")
        env_var = getattr(ENV, var)

        if not force and env_var in os.environ:
            print(f"Aldepyde variable {env_var} is already set. Use force=True to override")
            return

        os.environ[env_var] = str(val)
        print(f"Set {env_var} = {val}")
        return

    # TODO Test all this somehow
    @staticmethod
    def get_default_path():
        platform = sys.platform
        xdg = os.getenv('XDG_CACHE_HOME')
        if xdg:
            return os.path.join(os.path.expanduser(xdg), ENV.APP)
        if platform == "win32": # Windows
            base = os.getenv("LOCALAPPDATA", os.path.expanduser("~\\AppData\\Local"))
            return os.path.join(base, ENV.APP, "Cache")
        elif sys.platform == "darwin": # MacOS
            return os.path.join(os.path.expanduser("~/Library/Caches"), ENV.APP)
        else: # Linux without XDG set
            return os.path.join(os.path.expanduser("~/.cache"), ENV.APP)



from typing import List, Tuple, Optional, Dict
from dataclasses import dataclass, field
from ..biomolecule.structure import Atom, Residue, Chain, Model, Structure


@dataclass
class Loop():
    title: str
    canon_loop: bool
    tokens: List[str] = field(default_factory=list)
    # data: str = ""
    data: List[str] = field(default_factory=list)

    data_fields: List[list] = field(default_factory=list)

    def set_title(self, title):
        self.title = title

    def add_token(self, token):
        self.tokens.append(token)

    def add_data(self, data):
        # self.data += data
        self.data.append(data)

    def get_data(self) -> str:
        return "".join(self.data)

    def add_row(self, row):
        self.data_fields.append(row)

    def is_canon(self):
        return self.canon_loop

# Initial parse of loop sections
def catagorize_mmcif_sections(text):
    loops: Dict[str, Loop] = {}
    canon_loop = False
    in_text_field = False
    current_loop = Loop(str(None), False)
    for line in text.splitlines():
        # End of group
        if len(line.strip()) < 1:
            continue
        # TODO What happens if we have multiple data_ entries? Multiple structures can be in one file
        if line.lower().startswith('data_'):
            continue
        if line.strip() == '#':
            current_loop = Loop(str(None), False)
            continue
        if line.strip() == ';':
            in_text_field = False
        # Start of a canon loop
        if line.lower().startswith('loop_'):
            canon_loop = True
            continue
        line += '\n' # <-- I hate semicolons
        if canon_loop:
            title = line.strip()[:line.find('.')]
            current_loop = Loop(title, canon_loop)
            loops[title] = current_loop
            canon_loop = False
        if current_loop.is_canon() and line.startswith('_'):
            current_loop.add_token(line.strip())
            continue
        elif current_loop.is_canon() and not line.startswith('_'):
            current_loop.add_data(line)

        if not current_loop.is_canon():
            title = line.strip()[:line.find('.')]
            if title not in loops.keys():
                loops[title] = current_loop
                current_loop.set_title(title)
            current_loop.add_token(line.split()[0])
            current_loop.add_data(line[len(line.split()[0]):])
    return loops

def split_list(ls:list, n:int):
    q, r = divmod(len(ls), n)
    if r != 0:
        yield None
    try:
        for i in range(0, len(ls), n):
            yield ls[i:i + n]
    except ValueError:
        yield None

SPECIAL_CHARACTERS = {'\'', '\"', '\n;'}
def evaluate_fields(loops:Dict[str, Loop]):
    for key in loops.keys():
        loop = loops[key]
        current_value = ""
        special_flag = ""
        values = list()
        data = loop.get_data()
        for i in range(len(data)):
            char = data[i]
            if char in SPECIAL_CHARACTERS or data[i - 1]+char in SPECIAL_CHARACTERS:
                if special_flag == "":
                    special_flag = char
                elif special_flag == char:
                    special_flag = ""
            # Hit whitespace. Attempt to complete a value
            if char.isspace() and special_flag == "":
                if len(current_value) > 0:
                    values.append(current_value)
                    current_value = ""
                else:
                    continue
            else:
                current_value += char
        for l in split_list(values, len(loop.tokens)):
            loop.add_row(l)
    return loops
'''
    return Atom(serial=serial,
                name=atom_name,
                altloc=altLoc,
                coord=(x, y, z),
                occupancy=occupancy,
                b_factor=tempFactor,
                element=element,
                charge=charge,
                is_het=s.startswith('HETATM'))
'''
def _build_mmcif_structure(name:str, _atom_site: Loop) -> Structure|None:
    fl = _atom_site.tokens
    try:
        id_idx = fl.index('_atom_site.id') # Turns out this is the only required field. I require PDB minimum
        group_idx = fl.index('_atom_site.group_PDB')
        atom_name_idx = fl.index('_atom_site.auth_atom_id')
        x_idx = fl.index('_atom_site.Cartn_x')
        y_idx = fl.index('_atom_site.Cartn_y')
        z_idx = fl.index('_atom_site.Cartn_z')
        occ_idx = fl.index('_atom_site.occupancy')
        b_f_idx = fl.index('_atom_site.B_iso_or_equiv')
        res_name_idx = fl.index('_atom_site.auth_comp_id')
        res_num_idx = fl.index('_atom_site.auth_seq_id')
    except ValueError:
        return None
    try:
        i_code_idx = fl.index('_atom_site.pdbx_PDB_ins_code')
    except ValueError:
        i_code_idx = None
    try:
        model_idx = fl.index('_atom_site.pdbx_PDB_model_num')
    except ValueError:
        model_idx = None
    try:
        chain_id_idx = fl.index('_atom_site.auth_asym_id')
    except ValueError:
        chain_id_idx = None
    try:
        alt_loc_idx = fl.index('_atom_site.label_alt_id')
    except ValueError:
        alt_loc_idx = None
    try:
        element_idx = fl.index('_atom_site.type_symbol')
    except ValueError:
        element_idx = None
    try:
        charge_idx = fl.index('_atom_site.pdbx_formal_charge')
    except ValueError:
        charge_idx = None

    structure = Structure(name)
    models: Dict[int, Model] = {}
    chains: Dict[str, Chain] = {}
    residues: Dict[str, Residue] = {}
    current_model = Model('1')
    current_chain = Chain('A')

    if model_idx is None:
        structure.add_model(current_model)
    if charge_idx is None:
        current_model.add_chain(current_chain)

    for field in _atom_site.data_fields:
        atom = Atom(serial=field[id_idx],
             name=field[atom_name_idx],
             coord=(float(field[x_idx]), float(field[y_idx]), float(field[z_idx])),
             occupancy=float(field[occ_idx]),
             b_factor=float(field[b_f_idx]),
             altloc=field[alt_loc_idx] if alt_loc_idx is not None else None,
             element=field[element_idx] if element_idx is not None else None,
             charge=field[charge_idx] if charge_idx is not None else None,
             is_het=field[group_idx].startswith('HETATM'))
        res_id = field[res_num_idx]
        res_name = field[res_name_idx]
        i_code = field[i_code_idx] if i_code_idx is not None else None
        chain_id = field[chain_id_idx] if chain_id_idx is not None else 'A'
        model_id = field[model_idx] if chain_id_idx is not None else 1
        if model_id not in models:
            models[model_id] = Model(model_id)
            structure.add_model(models[model_id])
        current_model = models[model_id]
        if chain_id not in current_model.chains:
            chains[chain_id] = Chain(id=chain_id)
            current_model.add_chain(chains[chain_id])
        current_chain = models[model_id].get_chain(chain_id)
        if res_id not in current_chain.residues:
            residues[res_id] = Residue(res_name, res_id, i_code)
            chains[chain_id].add_residue(residues[res_id])
        residues[res_id].add_atom(atom) # Atoms should be guaranteed to have different _atom_site.id values
    return structure



def _parse_mmcif(text:str):
    loops = catagorize_mmcif_sections(text)
    loops = evaluate_fields(loops)
    try:
        structure_name = text[len('_data'):text.find('\n')].strip()
        structure = _build_mmcif_structure(structure_name, loops['_atom_site'])
        return structure
    except ZeroDivisionError:
        return None


from dataclasses import dataclass, field
from typing import Mapping

@dataclass(frozen=True)
class ProteinDefaults():
    alphabet: frozenset[str] = frozenset("ACDEFGHIKLMNPQRSTVWY")
    classes: Mapping[str, frozenset[str]] = field(default_factory=lambda:{
        "+":frozenset("RHK"),
        "-":frozenset("ED"),
        "polar":frozenset("NSCTQ"),
        "nonpolar":frozenset("MALIVGP"),
        "aromatic":frozenset("FYW"),
        "*":frozenset("")
    })
    flags = dict()

    def get_alphabet(self):
        return self.alphabet

    def get_classes(self):
        return self.classes



import io
from typing import Tuple
from aldepyde.databases._database import local_database
import operator
from contextlib import nullcontext
import re

class scop_parser(local_database):
    #TODO Do we want to handle custom classes?
    regex = re.compile(b">[a-zA-Z0-9_.]* *[a-l](.[0-9]+)?(.[0-9]+)?(.[0-9]+)?")


    op = {
        "and": lambda a,b,c: a and b and c,
        "or": lambda a,b,c: a or b or c
    }

    def fetch(self, url):
        pass

    def fetch_code(self, codes):
        pass

    def parse(self, text):
        pass

    def extract_all_scop(self):
        pass

    def partition_scope(self):
        pass

    def extract_all_astral(self):
        self.refresh()
        lines = self.fp.readlines()
        entry = b""
        for line in lines:
            if line.startswith(b">") and len(entry) > 0:
                yield entry
                entry = b""
            entry += line
        yield entry

    # TODO allow a list of search parameters. Big challenge to make efficient, but could be cute

    def yield_astral(self, class_name:str=b'',contains_id:str=b'' , contains_desc:str=b'', mode:str="and", include_newline:bool=False):
        class_name, contains_desc, contains_id = self._bytify(class_name, contains_desc, contains_id)
        mode = self._validate_mode(mode)
        logic = scop_parser.op[mode]
        for line in self.extract_all_astral():
            identifiers = self.regex.search(line).group().split()
            id = identifiers[0]
            cls = identifiers[1]
            unmatched_spl = self.regex.sub(b'', line).split(b'\n')
            desc = unmatched_spl[0]
            if include_newline:
                sequence = b"\n".join(unmatched_spl[1:])
            else:
                sequence = b"".join(unmatched_spl[1:])
            if logic(class_name.lower() in cls.lower(), contains_id.lower() in id.lower(), contains_desc.lower() in desc.lower()):
                yield sequence


    # TODO Create a new function that yields everything in the dictionary. Don't be hacky to avoid creating a file
    def partition_astral(self, destination:None|str=None, append=False, class_name:str=b'',contains_id:str=b'' , contains_desc:str=b'', mode="and") -> dict:
        class_name, contains_desc, contains_id = self._bytify(class_name, contains_desc, contains_id)
        mode = self._validate_mode(mode)
        logic = scop_parser.op[mode]

        if append:
            file_context = open(destination, 'ab') if destination is not None else nullcontext(io.BytesIO())
        else:
            file_context = open(destination, 'wb') if destination is not None else nullcontext(io.BytesIO())
        with file_context as fp:
            ret_dict = dict()
            for line in self.extract_all_astral():
                identifiers = self.regex.search(line).group().split()
                id = identifiers[0]
                cls = identifiers[1]
                unmatched_spl = self.regex.sub(b'', line).split(b'\n')
                desc = unmatched_spl[0]
                sequence = b"\n".join(unmatched_spl[1:])
                if logic(class_name.lower() in cls.lower(), contains_id.lower() in id.lower(), contains_desc.lower() in desc.lower()):
                    ret_dict[id] = { # Yes, I know '>' isn't part of the FASTA identifier. This keeps things more consistant
                        "class" : cls,
                        "description" : desc,
                        "sequence" : sequence
                    }
                    fp.write(line)
        return ret_dict

    def add_entry_to_dict(self, data:dict, id:str|bytes, sequence:str|bytes, cls:bytes|str|None=None, description:bytes|str=b''):
        id, sequence, cls, description = self._bytify(id, sequence, cls, description)
        data[b'>'+id] = {
                        "class" : cls,
                        "description" : description,
                        "sequence" : sequence
                    }
        return data

    def add_entry_to_target_file(self, filepath:str, id:bytes|str, sequence:bytes|str, cls:bytes|str|None=None, description:bytes|str=''):
        sequence, cls, description = self._bytify(sequence, cls, description)
        sequence = sequence + b'\n' if sequence[-1] != b'\n' else sequence
        entry = {
                    "class" : cls,
                    "description" : description,
                    "sequence" : sequence
                    }
        id_code = f'>{id}'.encode('utf-8')
        d = {id_code:entry}
        self.write_from_dict(d, destination=filepath, append=True)


    # Just simple dictionary union for now. Maybe later detect clashes?
    def combine_data(self, d1:dict, d2:dict) -> dict:
        return d1 | d2

    def add_label(self, data:dict, label:str, class_name:str=b'',
                  contains_id:str=b'' , contains_desc:str=b'', mode="and",
                  add_to_end:bool=False) -> dict:
        mode = self._validate_mode(mode)
        class_name, contains_desc, contains_id, label = self._bytify(class_name, contains_desc, contains_id, label)
        logic = scop_parser.op[mode]
        for key in data.keys():
            cls = data[key]['class']
            desc = data[key]['description']
            id = key
            if logic(class_name.lower() in cls.lower(), contains_id.lower() in id.lower(),
                     contains_desc.lower() in desc.lower()):
                if add_to_end:
                    data[key]['description'] = data[key]['description'] + label
                else:
                    data[key]['description'] = label + data[key]['description']
        return data

    def write_from_dict(self, data:dict, destination:str, append:bool=False):
        if append:
            fp = open(destination, 'ab')
        else:
            fp = open(destination, 'wb')
        for key in data.keys():
            w_str = key + b' ' + data[key]['class'] + b' ' + data[key]['description'] + b'\n' + data[key]['sequence']
            fp.write(w_str)
        fp.close()

    def _bytify(self, *args:str|bytes|None) -> Tuple[bytes, ...]:
        r_lst = list()
        for arg in args:
            if isinstance(arg, str):
                r_lst.append(arg.encode('utf-8'))
            elif arg is None:
                r_lst.append(None)
            else:
                r_lst.append(arg)
        return tuple(r_lst)

    def _validate_mode(self, mode):
        mode = mode.strip()
        if mode == "":
            mode = 'and'
        if mode != "and" and mode != "or":
            raise ValueError(f"Invalid mode: {mode}\n\tmode must be \"and\" or \"or\".")
        return mode.lower()

    #TODO Make this iterable?


from io import BytesIO

from aldepyde.databases.RemoteFileHandler import RemoteFileHandler
from aldepyde import get_cache

EXTENSIONS = {'pdb', 'mmcif'}
PDB_URL = "https://files.rcsb.org/pub/pdb/data/structures/divided/"

class pdb_Client():
    def attempt_cache(self, file) -> BytesIO|None:
        cache_path = get_cache().retrieve(file)
        if cache_path is None:
            return None
        else:
            with open(cache_path, 'rb') as fp:
                return BytesIO(fp.read())

    def fetch(self, code, extension='mmcif'):
        code = code.lower().strip()
        extension = extension.lower().strip()
        if extension == 'pdb':
            filepath = f"pdb/{code[1:3]}/pdb{code}.ent.gz"
            name = f"pdb{code}.ent.gz"
        elif extension == 'mmcif':
            filepath = f"mmCIF/{code[1:3]}/{code}.cif.gz"
            name = f"{code}.cif.gz"
        else:
            raise ValueError(f"Invalid file extension {extension}.\nAllowed values are {EXTENSIONS}\n")
        c_attempt = self.attempt_cache(name)
        if c_attempt is not None:
            return c_attempt
        fetch_url = PDB_URL + filepath
        data = RemoteFileHandler.fetch_url(fetch_url)
        get_cache().save_to_cache(data, name)
        return data




from typing import List, Tuple, Optional, Dict
from dataclasses import dataclass, field
import math
from aldepyde.data import get_store

from aldepyde.data import get_store

@dataclass
class Atom():
    serial: str
    name: str
    coord: Tuple[float, float, float]
    occupancy: float
    b_factor: float
    is_het: bool
    element: Optional[str]
    charge: Optional[str]
    altloc: Optional[str]
    pseudo: bool = False

    @property
    def location(self):
        return self.coord

    @property
    def x(self):
        return self.coord[0]

    @property
    def y(self):
        return self.coord[1]

    @property
    def z(self):
        return self.coord[2]


    def distance_to(self, other:'Atom') -> float:
        return math.sqrt((self.location[0] - other.location[0])**2 + (self.location[1] - other.location[1])**2+ (self.location[2] - other.location[2])**2)


@dataclass
class Residue():
    name: str
    id: str
    ins_code: Optional[str]
    atoms: List[Atom] = field(default_factory=list)
    # residue_type: str|None = None

    # def __post_init__(self):
    #     from aldepyde.Ops.residue_ops import residue_type
    #     self.residue_type = residue_type(self)

    @property
    def res_type(self):
        from aldepyde.Ops.residue_ops import residue_type
        return residue_type(self, self.is_het())

    def is_het(self) -> bool|None:
        if len(self.atoms) > 0:
            return self.atoms[0].is_het
        else:
            return None


    def add_atom(self, atom: Atom) -> None:
        self.atoms.append(atom)

    @property
    def one_letter(self) -> str:
        return get_store().one_letter()

    def dump_atom_labels(self, lower=False) -> List[str]:
        if lower:
            return [a.name.lower() for a in self.atoms]
        else:
            return [a.name for a in self.atoms]

    def get_atoms_by_label(self, labels:str|tuple) -> Tuple[Atom,...]:
        if isinstance(labels, str):
            labels = tuple(labels.split())
        atoms = list()
        for atom in self.atoms:
            if atom.name.lower() in labels:
                atoms.append(atom)
        return tuple(atoms)

    def centroid(self, as_atom=False, as_residue_type=None) -> Tuple[float, float, float]|Atom:
        from aldepyde.Ops.residue_ops import centroid
        if as_residue_type is not None:
            res_type = as_residue_type
        else:
            res_type = self.residue_type
        if not as_atom:
            return centroid(self, res_type)
        else:
            return Atom("-1", "Dum", centroid(self, res_type), occupancy=0, b_factor=0, is_het=True, element=None, charge=None, altloc=None)

    def center_of_mass(self, as_residue_type=None, as_atom=False):
        from aldepyde.Ops.residue_ops import center_of_mass
        if as_residue_type is not None:
            res_type = as_residue_type
        else:
            res_type = self.residue_type
        if not as_atom:
            return center_of_mass(self, res_type)
        else:
            return Atom("-1", "Dum", center_of_mass(self, res_type), occupancy=0, b_factor=0, is_het=True, element=None,
                        charge=None, altloc=None)

    def yield_atoms(self):
        for atom in self.atoms:
            yield atom

    def dump_atoms(self):
        return [a for a in self.yield_atoms()]

    def __str__(self):
        return f"{self.name}:{self.id}"

    def __repr__(self):
        return f"{self.name}:{self.id}"

@dataclass
class Chain():
    id: str
    residues: Dict[str, Residue] = field(default_factory=dict)

    def add_residue(self, residue: Residue):
        self.residues[residue.id] = residue
        # self.residues.append(residue)

    def get_residue(self, id):
        try:
            return self.residues[id]
        except KeyError:
            return self.residues[str(id)]

    def yield_atoms(self):
        for residue in self.residues.values():
            yield from residue.yield_atoms()

    def dump_atoms(self):
        return [a for a in self.yield_atoms()]

    def to_sequence(self) -> str:
        for res in self.residues:
            print(self.residues[res].res_type)

    def __str__(self):
        return self.id

    def __repr__(self):
        return self.id

@dataclass
class Model:
    id: str
    chains: Dict[str, Chain] = field(default_factory=dict)

    def add_chain(self, chain: Chain) -> None:
        self.chains[chain.id] = chain
        # self.chains.append(chain)

    def get_chain(self, id) -> Chain:
        try:
            return self.chains[id]
        except KeyError:
            return self.chains[str(id)]

    def yield_atoms(self):
        for chain in self.chains.values():
            yield from chain.yield_atoms()

    def dump_atoms(self):
        return [a for a in self.yield_atoms()]

    def __str__(self):
        return self.id

    def __repr__(self):
        return self.id

@dataclass
class Structure:
    id: str = "Unnamed Structure"
    # title: str = ""
    models: Dict[str, Model] = field(default_factory=dict)

    def add_model(self, model: Model) -> None:
        self.models[model.id] = model
        # self.models.append(model)

    def get_model(self, id) -> Model:
        try:
            return self.models[id]
        except KeyError:
            return self.models[str(id)]

    def yield_atoms(self):
        for model in self.models.values():
            yield from model.yield_atoms()

    def dump_atoms(self):
        return [a for a in self.yield_atoms()]

    def to_fasta(self) -> list[str]:
        return [x.to_fasta() for x in self.models]

    def __str__(self):
        return self.id

    def __repr__(self):
        return self.id


# import pkg_resources
# from aldepyde.remode_data import remode_data as remode_data

class ProteinStats:
    #TODO Clean this class a lot... also add some things
    #TODO Isoelectric point
    def __init__(self, His_Is_Charged=True, Cys_Is_Polar=True, Charged_Is_Polar=True):
        self.AAs = "ARNDCEQGHILKMFPSTWYV"
        self.POSITIVE = "RK"
        self.NEGATIVE = "ED"
        self.POLAR = "STNQ"
        self.NONPOLAR = "AVILMFYWGP"
        self.CHARGED = "RKED"

        if His_Is_Charged:
            self.CHARGED += "H"
            self.POSITIVE += "H"
        else:
            self.POLAR += "H"

        if Cys_Is_Polar:
            self.POLAR += "C"
        else:
            self.NONPOLAR += "C"

        if Charged_Is_Polar:
            self.POLAR += self.POSITIVE
            self.POLAR += self.NEGATIVE

    def GetNumResidues(self, sequence):
        ret_dic = {}
        for c in self.AAs:
            ret_dic[c] = sequence.count(c)
        return ret_dic


    def GetCharge(self, sequence):
        total_charge = 0
        for c in sequence:
            if c in self.POSITIVE:
                total_charge += 1
            elif c in self.NEGATIVE:
                total_charge -= 1
        return total_charge

    def NumCharged(self, sequence):
        total = 0
        for c in sequence:
            if c in self.CHARGED:
                total += 1
        return total

    def NumPolar(self, sequence):
        total = 0
        for c in sequence:
            if c in self.POLAR:
                total += 1
        return total

    def PercentCharged(self, sequence):
        num_charged = 0
        for c in sequence:
            if c in self.POSITIVE or c in self.NEGATIVE:
                num_charged += 1
        return num_charged / len(sequence)

    def PercentPolarC(self, sequence):
        num_polar = 0
        for c in sequence:
            if c in self.POLAR_CHECK:
                num_polar += 1
        return num_polar / len(sequence)

    def PercentPolar(self, sequence):
        num_polar = 0
        for c in sequence:
            if c in self.POLAR:
                num_polar += 1
        return num_polar / len(sequence)

    def PercentNonPolar(self, sequence):
        num_not = 0
        for c in sequence:
            if c in self.NONPOLAR:
                num_not += 1
        return num_not / len(sequence)

    # def GetMass(self, sequence, from_abundances=False):
    #     return remode_data.ProteinMass(sequence, from_abundances=from_abundances)